### Wrapper API

from metient.lib import vertex_labeling as vert_label
from metient.util import plotting_util as plutil
from metient.util import data_extraction_util as dutil
from metient.util import pairtree_data_extraction_util as ptutil

from metient.util.globals import *

def evaluate(T, ref, var, ordered_sites, primary_site, node_idx_to_label,
             weights, print_config, output_dir, run_name, G=None, O=None, batch_size=1024, 
             custom_colors=None, weight_init_primary=True, solve_polytomies=False):
    return vert_label.get_migration_history(T, ref, var, ordered_sites, primary_site, node_idx_to_label,
                                            weights, print_config, output_dir, run_name, G=G, O=O, 
                                            batch_size=batch_size, custom_colors=custom_colors, 
                                            weight_init_primary=weight_init_primary, mode='evaluate', solve_polytomies=solve_polytomies)

def calibrate(Ts, ref_matrices, var_matrices, ordered_sites, primary_sites, node_idx_to_labels,
              weights, print_config, output_dir, run_names, Gs=None, Os=None,
              batch_size=1024, custom_colors=None, weight_init_primary=True,  solve_polytomies=False):
    return vert_label.calibrate(Ts, ref_matrices, var_matrices, ordered_sites, primary_sites, node_idx_to_labels,
                                weights, print_config, output_dir, run_names, Gs=Gs, Os=Os, batch_size=batch_size, 
                                custom_colors=custom_colors, weight_init_primary=weight_init_primary, solve_polytomies=solve_polytomies)

class PrintConfig:
    def __init__(self, visualize=True, verbose=False, k_best_trees=3, save_outputs=True):
        '''
        visualize: bool, whether to visualize loss, best tree, and migration graph
        verbose: bool, whether to print debug info
        k_best_trees: int, number of best tree solutions to visualize (if 1, only show best tree)
        save_outputs: bool, whether to save pngs and pickle files 
        '''
        self.visualize = visualize
        self.verbose = verbose 
        self.k_best_trees = k_best_trees
        self.save_outputs = save_outputs

class Weights:
    def __init__(self,  mig=10.0, comig=5.0, seed_site=1.0, gen_dist=0.0, organotrop=0.0, data_fit=0.2, reg=2.0, entropy=0.1):
        if not isinstance(mig, list):
            mig = [mig]
        if not isinstance(seed_site, list):
            seed_site = [seed_site]
        self.data_fit = data_fit
        self.mig = mig
        self.comig = comig
        self.seed_site = seed_site
        self.reg = reg
        self.gen_dist = gen_dist
        self.organotrop = organotrop
        self.entropy = entropy


########### Data Extraction Utilities ############

# TODO: this API stucture is confusing when you have multiple trees per patient
def get_ref_var_matrices(tsv_filepaths, split_char=None):
    '''
    Args:
        - tsv_filepaths: either a single path or list of paths to tsvs (one for each patient). Expects columns:

        ['sample_index', 'sample_label', 'anatomical_site_index',
        'anatomical_site_label', 'character_index', 'character_label', 'ref', 'var']
        - split_char: if clustering was used, indicates how the mutations in the 
        character_label are split by. e.g. "_". If passed, we will shorten the character_label
        in the dictionary mapping index to character_label that we return for visualization purposes

    Returns:
        If a list of tsv_filepaths is supplied, a list of each of the following is returned 
        (otherwise, a single item of each is returned)
        (1) R matrix (num_samples x num_clusters) with the # of reference reads for each sample+cluster,
        (2) V matrix (num_samples x num_clusters) with the # of variant reads for each sample+cluster,
        (3) list of unique anatomical sites from the patient's data,
        (4) dictionary mapping index to character_label (based on input tsv, gives the index for each mutation name,
        where these indices are used in R matrix, V matrix

    For each of the returned lists, index 0 represents the patient at the first tsv_filepath, etc.
    '''
    return dutil.get_ref_var_matrices(tsv_filepaths, split_char=split_char)

def get_adj_matrices_from_pairtree_results(pairtee_results_filepaths):
    '''
    Args:
        - pairtee_results_filepaths: either a single path or list of paths to pairtree/orchard clone tree
        results (end with ".results.npz")

    Returns:
        If a list of pairtee_results_filepaths is supplied, a list of adjacency matrices is returned
        (otherwise, a single adjacency matrix is returned)

    '''
    return ptutil.get_adj_matrices_from_pairtree_results(pairtee_results_filepaths)

def get_genetic_distance_matrices_from_adj_matrices(adj_matrices, idx_to_character_labels, split_char, normalize=True):
    '''
    Get the genetic distances between nodes by counting the number of mutations between parent and child. 
    
    Args:
        - adj_matrices: either a single adjacency matrix or list of adjacency matrices
        - idx_to_character_labels: either a single dictionary or list of dictionaries.
        idx_to_character_label's keys are expected to be mutation/cluster indices with
        the mutation or mutations in that cluster (e.g. 'ENAM:4:71507837_DLG1:3:196793590'). 
        - split_char: indicates what the mutations in the cluster name are split by. e.g. "_". If
        split_char is None, it is assumed that these are single mutations and not clusters, and the distance
        between each parent and child node in the clone tree is 1. 

    If a single adj_matrix or idx_to_character_label is inputted, a single genetic distance matrix is returned. 
    Otherwise, a list of genetic distance matrices is returned.
    '''
    return dutil.get_genetic_distance_matrices_from_adj_matrices(adj_matrices, idx_to_character_labels, split_char, normalize=normalize)

def get_organotropism_matrix_from_msk_met(ordered_sites, cancer_type, frequency_csv, site_to_msk_met_map=None):
    '''
    Args:
        - ordered_sites: array of the anatomical site names (e.g. ["breast", "lung"])
        with length =  num_anatomical_sites) where the order matches the order of sites
        in the ref_matrix and var_matrix
        - cancer_type: cancer type name, which is used as the 
        - frequency_csv: csv with frequency of metastasis by cancer type
        - site_to_msk_met_map: dictionary mapping site names to MSK-MET site names. if
        not provided, the names used in the ordered_sites array are used

    Returns:
        array of size len(ordered_sites) with the frequency with which
        primary cancer type seeds site i
    '''
    return dutil.get_organotropism_matrix_from_msk_met(ordered_sites, cancer_type, frequency_csv, site_to_msk_met_map=site_to_msk_met_map)

def get_adj_matrices_from_spruce_mutation_trees(mut_trees_filename, idx_to_character_label, is_sim_data=False):
    '''
    When running MACHINA's generatemutationtrees executable, it provides a txt file with
    all possible mutation trees. See data/machina_simulated_data/mut_trees_m5/ for examples

    Returns a list of tuples, each containing (T, pruned_idx_to_character_label) for each
    tree in mut_trees_filename.
        - T: adjacency matrix where Tij = 1 if there is a path from i to j
        - idx_to_character_label: a dict mapping indices of the adj matrix T to character
        labels 
    '''
    return dutil.get_adj_matrices_from_spruce_mutation_trees(mut_trees_filename, idx_to_character_label, is_sim_data=is_sim_data)

def get_idx_to_cluster_label(cluster_filepath, ignore_polytomies):
    '''
    Args:
	    cluster_filepath: path to cluster file for MACHINA simulated data in the format:
	    0
	    1
	    3;15;17;22;24;29;32;34;53;56
	    69;78;80;81

    	where each semi-colon separated number represents a mutation that belongs
    	to cluster i where i is the file line number.

	    ignore_polytomies: whether to include resolved polytomies (which were found by
	    running PMH-TR) in the returned dictionary

    Returns:
        (1) a dictionary mapping cluster number to cluster name
        for e.g. for the file above, this would return:
            {0: '0', 1: '69;78;80;81'}
    '''
    return dutil.get_idx_to_cluster_label(cluster_filepath, ignore_polytomies)

def get_ref_var_matrices_from_machina_sim_data(tsv_filepath, pruned_idx_to_cluster_label, T):
    '''
    tsv_filepath: path to tsv for machina simulated data (generated from create_conf_intervals_from_reads.py)

    tsv is expected to have columns: ['#sample_index', 'sample_label', 'anatomical_site_index',
    'anatomical_site_label', 'character_index', 'character_label', 'f_lb', 'f_ub', 'ref', 'var']

    pruned_idx_to_cluster_label:  dictionary mapping the cluster index to label, where 
    index corresponds to col index in the R matrix and V matrix returned. This isn't 1:1 
    with the 'character_label' to 'character_index' mapping in the tsv because we only keep the
    nodes which appear in the mutation tree, and re-index after removing unseen nodes
    (see _get_adj_matrix_from_machina_tree)

    T: adjacency matrix of the internal nodes.

    returns
    (1) R matrix (num_samples x num_clusters) with the # of reference reads for each sample+cluster,
    (2) V matrix (num_samples x num_clusters) with the # of variant reads for each sample+cluster,
    (3) unique anatomical sites from the patient's data
    '''
    return dutil.get_ref_var_matrices_from_machina_sim_data(tsv_filepath, pruned_idx_to_cluster_label, T)



import torch
from src.lib import vertex_labeling
from src.util import data_extraction_util as data_util
from src.util import pairtree_data_extraction_util as pt_util
from src.util import vertex_labeling_util as vert_util
from src.util import plotting_util as plot_util
import os

cd = os.getcwd()
os.chdir(cd)
data_dir = os.path.join(cd, "data/gundem_neuroblastoma_2023/patient_driver_genes/")
output_dir = os.path.join(cd, "metient_example_outputs/")
if not os.path.exists(output_dir):
   os.makedirs(output_dir)
patient_id = "H103207_clustered"

# Process input reference and variant reads
ref_var_fn = os.path.join(data_dir, f"{patient_id}_SNVs.tsv")
ref_matrix, var_matrix, unique_sites, idx_to_cluster_label = data_util.get_ref_var_matrices_from_real_data(ref_var_fn)

# Process input trees (this is for PairTree or Orchard trees)
tree_data = pt_util.get_adj_matrices_from_pairtree_results(os.path.join(data_dir, "orchard_trees", f"{patient_id}.results.npz"))

# Specify anatomical site label of primary site
primary_site = "left adrenal - D"

# Enumerate all possible trees (if there are multiple) and run migration history analysis
for i, (T, llh) in enumerate(tree_data):
    print(f"Solving migration history for tree {i}")
    if not vert_util.is_tree(T):
        print("Invalid tree was provided, skipping: \n", T)
        continue

    # Configure this to diplay mutations in a custom way 
    idx_to_cluster_label = {k:(" ").join(v.split('_')[0:2]) for k,v in idx_to_cluster_label.items()}
    primary_idx = unique_sites.index(primary_site)
    p = torch.nn.functional.one_hot(torch.tensor([primary_idx]), num_classes=len(unique_sites)).T
    
    weights = vertex_labeling.Weights(data_fit=1.0, mig=3.0, comig=2.0, seed_site=1.0, reg=2.0, gen_dist=0.0)
    print_config = plot_util.PrintConfig(visualize=False, verbose=False, viz_intermeds=False, k_best_trees=4)
    vertex_labeling.get_migration_history(T, ref_matrix, var_matrix, unique_sites, p, idx_to_cluster_label, weights,
                                          print_config, output_dir, patient_id, batch_size=64, max_iter=200)




import numpy as np
import torch
import networkx as nx
import math
from queue import Queue

from metient.util.globals import *

import scipy.sparse as sp
import pandas as pd
pd.options.display.float_format = '{:,.3f}'.format

if torch.cuda.is_available():
    torch.set_default_tensor_type(torch.cuda.FloatTensor)

class LabeledTree:
    # TODO: remove tree from here
    def __init__(self, tree, labeling):
        if (tree.shape[0] != tree.shape[1]):
            raise ValueError("Adjacency matrix should have shape (num_nodes x num_nodes)")
        if (tree.shape[0] != labeling.shape[1]):
            raise ValueError("Vertex labeling matrix should have shape (num_sites x num_nodes)")

        self.tree = tree
        self.labeling = labeling

    def __eq__(self, other):
        return ( isinstance(other, LabeledTree) and 
                (torch.equal(torch.nonzero(self.labeling), torch.nonzero(other.labeling))) and
                (torch.equal(torch.nonzero(self.tree), torch.nonzero(other.tree))))

    def __hash__(self):
        return hash(self.labeling.numpy().tobytes())*hash(self.tree.numpy().tobytes())

    def __str__(self):
        A = str(np.where(self.tree == 1))
        V = str(np.where(self.labeling == 1))
        return f"Tree: {A}\nVertex Labeling: {V}"


def tree_iterator(T):
    '''
    iterate an adjacency matrix, returning i and j for all values = 1
    '''
    # enumerating through a torch tensor is pretty computationally expensive,
    # so convert to a sparse matrix to efficiently access non-zero values
    T = T if isinstance(T, np.ndarray) else T.detach().numpy()
    T = sp.coo_matrix(T)
    for i, j, val in zip(T.row, T.col, T.data):
        yield i,j


def has_leaf_node(adjacency_matrix, start_node, num_internal_nodes):
    num_nodes = adjacency_matrix.size(0)
    visited = torch.zeros(num_nodes, dtype=torch.bool)
    queue = Queue()

    queue.put(start_node)
    visited[start_node] = True

    while not queue.empty():
        current_node = int(queue.get())
        # Get neighbors of the current node
        neighbors = torch.nonzero(adjacency_matrix[current_node]).squeeze(1)

        # Enqueue unvisited neighbors
        for neighbor in neighbors:
            neighbor = int(neighbor)
            if not visited[neighbor]:
                queue.put(neighbor)
                visited[neighbor] = True
                # it's a leaf node if its not in the clone tree idx label dict
                if neighbor > num_internal_nodes:
                    return True
    return False

def get_root_index(T):
    '''
    returns the root idx (node with no inbound edges) from adjacency matrix T
    '''

    candidates = set([x for x in range(len(T))])
    for i, j in tree_iterator(T):
        candidates.remove(j)
    msg = "More than one" if len(candidates) > 1 else "No"
    assert (len(candidates) == 1), f"{msg} root node detected"

    return list(candidates)[0]

def restructure_matrices(adj_matrix,ref_matrix, var_matrix, node_idx_to_label, gen_dist_matrix):
    """
    Restructure the adjacency matrix, ref matrix and var matrix so that the node at index 0 becomes the root node.

    Parameters:
    - adj_matrix: 2D tensor representing the adjacency matrix.

    Returns:
    - Restructured adjacency matrix.
    """
    root_idx = get_root_index(adj_matrix)
    new_order = [x for x in range(len(adj_matrix)) if x != root_idx]
    new_order.insert(0, root_idx)

    # Swap rows and columns to move the first row and column to the desired position
    swapped_adjacency_matrix = adj_matrix[new_order, :][:, new_order]
    swapped_ref_matrix = ref_matrix[:, new_order]
    swapped_var_matrix = var_matrix[:, new_order]
    if gen_dist_matrix == None:
        swapped_gen_dist_matrix = None
    else:
        swapped_gen_dist_matrix = gen_dist_matrix[new_order, :][:, new_order]

    original_root_label = node_idx_to_label[root_idx]
    original_node0_label = node_idx_to_label[0]
    node_idx_to_label[0] = original_root_label
    node_idx_to_label[root_idx] = original_node0_label

    return swapped_adjacency_matrix, swapped_ref_matrix, swapped_var_matrix, node_idx_to_label, swapped_gen_dist_matrix


def top_k_integers_by_count(lst, k, thres, cutoff):
    # find unique sites and their counts
    unique_sites, counts = np.unique(lst, return_counts=True)
    # filter for sites that occur at least thres times
    filtered_values = unique_sites[counts >= thres]
    # recount after filtering
    unique_sites, counts = np.unique(filtered_values, return_counts=True)
    # sort based on counts in descending order
    sorted_indices = np.argsort(counts)[::-1]
    # get the top k integers and their counts
    if len(unique_sites) > k and cutoff:
        return list(unique_sites[sorted_indices[:k]])
        
    return list(unique_sites)

def get_resolver_sites_of_node_and_children(input_T, T, U, num_children, node_idx, thres, cutoff):
    '''
    Looking at the sites that node_idx and node_idx's children are in,
    figure out the sites that the resolver nodes should be initialized with
    (any sites that are detected >= 2 times)
    '''
    clone_tree_children = input_T[node_idx].nonzero()[:,0].tolist()
    # calculated based on clone tree children and leaf 
    # nodes estimated form U
    num_possible_resolvers = math.floor(num_children/2)
    U = U[:,1:] # don't include column for normal cells
    # get the sites that the node and node's children are in
    node_and_children = clone_tree_children+ [int(node_idx)]
    sites = (U[:,node_and_children] > U_CUTOFF).nonzero()[:,0]
    return top_k_integers_by_count(sites, num_possible_resolvers, thres, cutoff)
    
def get_k_or_more_children_nodes(input_T, T, U, k, thres, cutoff=True):
    '''
    returns the indices and proposed sites for resolver nodes
    that are placed under existing nodes with k or more children
    e.g. 
    input_T =[[0,1,1,],
             [0,0,0],
             [0,0,0],]
    T = [[0,1,1,1,0,0,0],
         [0,0,0,0,1,0,1],
         [0,0,0,0,0,1,0],
         [0,0,0,0,0,0,0],
         [0,0,0,0,0,0,0],
         [0,0,0,0,0,0,0],
         [0,0,0,0,0,0,0],]
    U = [[0,0,1,0,],
         [0,1,1,1,],]
    k = 3
    returns ([0], [[1]]) (node at index 0 has 3 children, and the resolver node under it should be
    place in site 1, since node 0 and its child node 1 are both detected in site 1)
    '''
    row_sums = torch.sum(T, axis=1)
    node_indices = list(np.where(row_sums >= k)[0])
    filtered_node_indices = []
    all_resolver_sites = []
    num_children = []
    for node_idx in node_indices:
        # get the sites that node_idx and node_idx's children are estimated in U
        resolver_sites = get_resolver_sites_of_node_and_children(input_T, T, U, row_sums[node_idx], node_idx, thres, cutoff)
        # a resolver node wouldn't help for this polytomy
        if len(resolver_sites) == 0:
            continue
        filtered_node_indices.append(node_idx)
        all_resolver_sites.append(resolver_sites)
    return filtered_node_indices, all_resolver_sites

# def get_k_or_more_children_nodes(T, k):
#     '''
#     returns the indices of nodes with k or more children
#     '''
#     row_sums = torch.sum(T, axis=1)
#     return list(np.where(row_sums >= k)[0])

def get_child_indices(T, indices):
    '''
    returns the indices of direct children of the nodes at indices
    '''
    all_child_indices = []

    for parent_idx in indices:
        children = np.where(T[parent_idx,:] > 0)[0]
        for child_idx in children:
            if child_idx not in all_child_indices:
                all_child_indices.append(child_idx)

    return all_child_indices

def repeat_n(x, n):
    '''
    Repeats tensor x 'n' times along the first axis, returning a tensor
    w/ dim (n, x.shape[0], x.shape[1])
    '''
    if n == 0:
        return x
    return x.repeat(n,1).reshape(n, x.shape[0], x.shape[1])

def get_path_matrix(T, remove_self_loops=False):
    bs = 0 if len(T.shape) == 2 else T.shape[0]
    # Path matrix that tells us if path exists from node i to node j
    I = repeat_n(torch.eye(T.shape[1]), bs)
    # M is T with self loops.
    # Convert to bool to get more efficient matrix multiplicaton
    B = torch.logical_or(T,I).int()
    # Implementing Algorithm 1 here, which uses repeated squaring to efficiently calc path matrix:
    # https://courses.grainger.illinois.edu/cs598cci/sp2020/LectureNotes/lecture1.pdf
    k = np.ceil(np.log2(T.shape[1]))
    for i in range(int(k)):
        B = torch.matmul(B, B)
    if remove_self_loops:
        B = torch.logical_xor(B,I)
    P = torch.sigmoid(BINARY_ALPHA * (2*B - 1))
    return P

# def adj_matrix_hash(T):
#     sparse_T = sp.coo_matrix(T.detach().numpy())
#     return hash((str(sparse_T.row), str(sparse_T.col)))

# def get_single_adj_matrix_path_matrix(T, remove_self_loops):
#     # Path matrix that tells us if path exists from node i to node j
#     I = torch.eye(T.shape[1])
#     # M is T with self loops.
#     # Convert to bool to get more efficient matrix multiplicaton
#     B = torch.logical_or(T,I).int()
#     # Implementing Algorithm 1 here, which uses repeated squaring to efficiently calc path matrix:
#     # https://courses.grainger.illinois.edu/cs598cci/sp2020/LectureNotes/lecture1.pdf
#     k = np.ceil(np.log2(T.shape[1]))
    
#     for i in range(int(k)):
#         B = torch.matmul(B, B)
#     if remove_self_loops:
#         B = torch.logical_xor(B,I)
#     P = torch.sigmoid(BINARY_ALPHA * (2*B - 1))
#     return P
    
# def get_path_matrix(T, remove_self_loops=False):
#     # no need to do any caching stuff when it's just the singular adjacency matrix
#     if len(T.shape) == 2:
#         T = T.reshape(1, T.shape[0], T.shape[1])
#         print(T.shape)
    
#     num_times_found = 0
#     global PATH_MATRICES_CACHE
#     final_P = torch.zeros(T.shape)
#     for bs_idx in range(T.shape[0]):
#         hsh = adj_matrix_hash(T[bs_idx])
#         if hsh in PATH_MATRICES_CACHE:
#             final_P[bs_idx] = PATH_MATRICES_CACHE[hsh]
#             num_times_found += 1
#         else:
#             P = get_single_adj_matrix_path_matrix(T[bs_idx], remove_self_loops)
#             final_P[bs_idx] = P
#             PATH_MATRICES_CACHE[hsh] = P
#     return final_P

def get_path_matrix_tensor(A):
    '''
    A is a numpy adjacency matrix
    '''
    return torch.tensor(get_path_matrix(A, remove_self_loops=True), dtype = torch.float32)

def get_mutation_matrix_tensor(A):
    '''
    A is an numpy ndarray or tensor adjacency matrix (where Aij = 1 if there is a path from i to j)

    returns a mutation matrix B, which is a subclone x mutation binary matrix, where Bij = 1
    if subclone i has mutation j.
    '''
    return torch.tensor(get_path_matrix(A.T, remove_self_loops=False), dtype = torch.float32)

def get_adj_matrix_from_edge_list(edge_list):
    T = []
    nodes = set([node for edge in edge_list for node in edge])
    T = [[0 for _ in range(len(nodes))] for _ in range(len(nodes))]
    for edge in edge_list:
        T[ord(edge[0]) - 65][ord(edge[1]) - 65] = 1
    return torch.tensor(T, dtype = torch.float32)


def is_tree(adj_matrix):
    rows, cols = np.where(adj_matrix == 1)
    edges = zip(rows.tolist(), cols.tolist())
    g = nx.Graph()
    g.add_edges_from(edges)
    return (not nx.is_empty(g) and nx.is_tree(g))


import torch
import sys
import numpy as np
import datetime
from tqdm import tqdm
import heapq
import torchopt

from metient import metient as met
from metient.util import vertex_labeling_util as vutil
from metient.util import eval_util as eutil
from metient.util import plotting_util as plot_util
import torch.optim.lr_scheduler as lr_scheduler
from metient.util.globals import *

from torch.distributions.binomial import Binomial

import os
import csv
import copy
import shutil
import pickle
import gzip
import json

G_IDENTICAL_CLONE_VALUE = None

# gumbel noise is initialized non-randomly for latent variables
RANDOM_VALS = None
LAST_P = None
SOLVE_POLYS = False

if torch.cuda.is_available():
    torch.set_default_tensor_type(torch.cuda.FloatTensor)

def add_batch_dim(x):
    return x.reshape(1, x.shape[0], x.shape[1])

def get_ancestral_labeling_metrics(V, A, G, O, p, i=-1, max_iter=-1):
    single_A = A
    bs = V.shape[0]
    num_sites = V.shape[1]
    num_nodes = V.shape[2]
    A = A if len(A.shape) == 3 else vutil.repeat_n(single_A, bs) 

    # 1. Migration number
    VA = V @ A
    VT = torch.transpose(V, 2, 1)
    site_adj = VA @ VT
    site_adj_trace = torch.diagonal(site_adj, offset=0, dim1=1, dim2=2).sum(dim=1)
    m = torch.sum(site_adj, dim=(1, 2)) - site_adj_trace

    # 2. Seeding site number
    # remove the same site transitions from the site adj matrix
    site_adj_no_diag = torch.mul(site_adj, vutil.repeat_n(1-torch.eye(num_sites, num_sites), bs))
    row_sums_site_adj = torch.sum(site_adj_no_diag, axis=2)
    # can only have a max of 1 for each site (it's either a seeding site or it's not)
    
    binarized_row_sums_site_adj = torch.sigmoid(BINARY_ALPHA * (2*row_sums_site_adj - 1)) # sigmoid for soft thresholding
    s = torch.sum(binarized_row_sums_site_adj, dim=(1))

    # 3. Comigration number
    VAT = torch.transpose(VA, 2, 1)
    W = VAT @ VA # 1 if two nodes' parents are the same color
    X = VT @ V # 1 if two nodes are the same color
    Y = torch.sum(torch.mul(VAT, 1-VT), axis=2) # Y has a 1 for every node where its parent has a diff color
    shared_par_and_self_color = torch.mul(W, X) # 1 if two nodes' parents are same color AND nodes are same color
    # tells us if two nodes are (1) in the same site and (2) have parents in the same site
    # and (3) there's a path from node i to node j
    global LAST_P # this is expensive to compute, so hash it if we don't need to update it
    if LAST_P != None and not update_adj_matrix(i, max_iter):
        P = LAST_P
    else:
        P = vutil.get_path_matrix_tensor(A)
        LAST_P = P

    shared_path_and_par_and_self_color = torch.sum(torch.mul(P, shared_par_and_self_color), axis=2)
    repeated_temporal_migrations = torch.sum(torch.mul(shared_path_and_par_and_self_color, Y), axis=1)
    binarized_site_adj = torch.sigmoid(BINARY_ALPHA * (2 * site_adj - 1))
    bin_site_trace = torch.diagonal(binarized_site_adj, offset=0, dim1=1, dim2=2).sum(dim=1)
    c = torch.sum(binarized_site_adj, dim=(1,2)) - bin_site_trace + repeated_temporal_migrations

    # 4. Genetic distance
    g = 0
    if G != None:
        #print("G\n", G[0,:,:])
        # calculate if 2 nodes are in diff sites and there's an edge between them (i.e there is a migration edge)
        R = torch.mul(A, (1-X))
        adjusted_G = -torch.log(G+0.01)
        R = torch.mul(R, adjusted_G)
        # if bs == 1:
        #     nonzero_indices =  torch.nonzero(R[0])
        #     print("gen dist")
        #     print("R\n", R)
        #     print(nonzero_indices)
        #     print(R[0][nonzero_indices])

        # TODO: get the average of genetic distance scores or normalize by max?
        g = torch.sum(R, dim=(1,2))/(m)

    # 5. Organotropism
    o = 0
    if O != None:
        # the organotropism frequencies can only be used on the first 
        # row, which is for the migrations from primary cancer site to
        # other metastatic sites (we don't have frequencies for every 
        # site to site migration)
        prim_site_idx = torch.nonzero(p)[0][0]
        O = O.repeat(bs,1).reshape(bs, O.shape[0])
        adjusted_freqs = -torch.log(O+0.01)
        organ_penalty = torch.mul(site_adj_no_diag[:,prim_site_idx,:], adjusted_freqs)
        #print(organ_penalty[0])
        o = torch.sum(organ_penalty, dim=(1))/(num_sites-1)

    return m, c, s, g, o

def get_entropy(V, soft_V):
    return -torch.sum(torch.mul(soft_V, torch.log2(soft_V)), dim=(1, 2)) / V.shape[2]

def get_repeating_weight_vector(bs, weight_list):
    repeat_times = bs // len(weight_list)

    # Repeat each number in the list the calculated number of times
    weights_vec = torch.tensor(weight_list * repeat_times)

    # If n is not a multiple of the length of the list, repeat the elements as necessary
    remaining_elements = bs % len(weight_list)
    if remaining_elements > 0:
        weights_vec = torch.cat([weights_vec, torch.tensor(weight_list[:remaining_elements])])

    return weights_vec

def get_mig_weight_vector(bs, weights):
    return get_repeating_weight_vector(bs, weights.mig)

def get_seed_site_weight_vector(bs, weights):
    return get_repeating_weight_vector(bs, weights.seed_site)

def ancestral_labeling_objective(V, soft_V, A, G, O, p, weights, i=-1, max_iter=-1):
    '''
    Args:
        V: Vertex labeling of the full tree (batch_size x num_sites x num_nodes)
        A: Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)
        G: Matrix of genetic distances between internal nodes (shape:  num_internal_nodes x num_internal_nodes).
        Lower values indicate lower branch lengths, i.e. more genetically similar.
        O: Array of frequencies with which the primary cancer type seeds site i (shape: num_anatomical_sites).
        p: one-hot vector indicating site of the primary
        weights: Weights object

    Returns:
        Loss to score the ancestral vertex labeling of the given tree. This combines (1) migration number, (2) seeding site
        number, (3) comigration number, and optionally (4) genetic distance and (5) organotropism.
    '''
    bs = V.shape[0]
    
    #assert(A.shape[0]==A.shape[1]==V.shape[2])
    m, c, s, g, o = get_ancestral_labeling_metrics(V, A, G, O, p, i, max_iter)

    # Entropy
    e = get_entropy(V, soft_V)

    num_sites, num_nodes = V.shape[1], V.shape[2]

    # TODO: should we normalize??
    #max_comigrations = num_sites**2
    # normalized_migrations = m/(num_nodes - 1)
    # normalized_comigrations = c/max_comigrations
    # normalized_seeding_sites = s/(num_sites)

    # normalized_migrations = m
    # normalized_comigrations = c
    # normalized_seeding_sites = s

    # normed_averaged_migrations = (weights.mig_delta*normalized_migrations) + ((1-weights.mig_delta)*normalized_comigrations)

    # averaged_migrations = (weights.mig_delta*m) + ((1-weights.mig_delta)*c)

    # Combine all 5 components with their weights
    # Explore different weightings
    if isinstance(weights.mig, list) and isinstance(weights.seed_site, list):
        mig_weights_vec = get_mig_weight_vector(bs, weights)
        seeding_sites_weights_vec = get_seed_site_weight_vector(bs, weights)
        mig_loss = torch.mul(mig_weights_vec, m)
        seeding_loss = torch.mul(seeding_sites_weights_vec, s)
        vertex_labeling_loss = (mig_loss + weights.comig*c + seeding_loss + weights.gen_dist*g + weights.organotrop*o+ weights.entropy*e)
        
    else:
        mig_loss = weights.mig*m
        seeding_loss = weights.seed_site*s
        vertex_labeling_loss = (mig_loss + weights.comig*c + seeding_loss + weights.gen_dist*g + weights.organotrop*o+ weights.entropy*e)

    loss_components = {MIG_KEY: m, COMIG_KEY:c, SEEDING_KEY: s, ORGANOTROP_KEY: o, GEN_DIST_KEY: g, ENTROPY_KEY: e}

    return vertex_labeling_loss, loss_components

# Adapted from PairTree
def calc_llh(F_hat, R, V, omega_v, epsilon=1e-5):
    '''
    Args:
        F_hat: estimated subclonal frequency matrix (num_nodes x num_mutation_clusters)
        R: Refernce matrix (num_samples x num_mutation_clusters)
        V: Variant matrix (num_samples x num_mutation_clusters)
    Returns:
        Data fit using the Binomial likelihood (p(x|F_hat)). See PairTree
        supplement section 2.2 for details.
    '''

    N = R + V
    S, K = F_hat.shape

    for matrix in V, N, omega_v:
        assert(matrix.shape == (S, K-1))

    # TODO: how do we make sure the non-cancerous root clone subclonal frequencies here are 1
    #assert(np.allclose(1, phi[0]))
    P = torch.mul(omega_v, F_hat[:,1:])

    # TODO: why these cutoffs?
    #P = torch.maximum(P, epsilon)
    #P = torch.minimum(P, 1 - epsilon)

    bin_dist = Binomial(N, P)
    F_llh = bin_dist.log_prob(V)
    #phi_llh = stats.binom.logpmf(V, N, P) / np.log(2)
    assert(not torch.any(F_llh.isnan()))
    assert(not torch.any(F_llh.isinf()))

    # TODO: why the division by K-1 and S?
    llh_per_sample = -torch.sum(F_llh, axis=1) / (K-1)
    nlglh = torch.sum(llh_per_sample) / S
    return (F_llh, llh_per_sample, nlglh)

def subclonal_presence_objective(ref, var, omega_v, U, B, weights):
    '''
    Args:
        ref: Reference matrix (num_anatomical_sites x num_mutation_clusters). Num. reads that map to reference allele
        var: Variant matrix (num_anatomical_sites x num_mutation_clusters). Num. reads that map to variant allele
        U: Mixture matrix (num_sites x num_internal_nodes)
        B: Mutation matrix (shape: num_internal_nodes x num_mutation_clusters)
        weights: Weights object

    Returns:
        Loss to score the ancestral vertex labeling of the given tree. This combines (1) migration number, (2) seeding site
        number, (3) comigration number, and optionally (4) genetic distance and (5) organotropism.
    '''
    # 1. Data fit
    F_hat = (U @ B)
    F_llh, llh_per_sample, nlglh = calc_llh(F_hat, ref, var, omega_v)

    # 2. Regularization to make some values of U -> 0
    reg = torch.sum(U)

    subclonal_presence_loss = (weights.data_fit*nlglh + weights.reg*reg)
    loss_components = {DATA_FIT_KEY: round(nlglh.item(), 3), REG_KEY: reg.item()}
    return subclonal_presence_loss, loss_components

def no_cna_omega(shape):
    '''
    Returns omega values assuming no copy number alterations (0.5)
    Shape is (num_anatomical_sites x num_mutation_clusters)
    '''
    # TODO: don't hardcode omega here (omegas are all 1/2 since we're assuming there are no CNAs)
    return torch.ones(shape) * 0.5

def get_random_vals_fixed_seeds(shape):
    global RANDOM_VALS
    if RANDOM_VALS != None and shape in RANDOM_VALS:
        return RANDOM_VALS[shape]

    if RANDOM_VALS == None:
        RANDOM_VALS = dict()

    rands = torch.zeros(shape)
    for i in range(shape[0]):
        torch.manual_seed(i)
        rands[i] = torch.rand(shape[1:])
    RANDOM_VALS[shape] = rands
    return RANDOM_VALS[shape]

def sample_gumbel(shape, eps=1e-20):
    G = get_random_vals_fixed_seeds(shape)
    return -torch.log(-torch.log(G + eps) + eps)

def gumbel_softmax_sample(logits, temperature):
    y = logits + sample_gumbel(logits.size())
    return torch.nn.functional.softmax(y / temperature, dim=1)

def gumbel_softmax(logits, temperature, hard=False):
    '''
    Adapted from https://discuss.pytorch.org/t/stop-gradients-for-st-gumbel-softmax/530/5

    Sample from the Gumbel-Softmax distribution and optionally discretize.
    Args:
        logits: [batch_size, n_class] unnormalized log-probs
        temperature: non-negative scalar
        hard: if True, take argmax, but differentiate w.r.t. soft sample y
    Returns:
        [batch_size, n_class] sample from the Gumbel-Softmax distribution.
        If hard=True, then the returned sample will be one-hot, otherwise it will
        be a probabilitiy distribution that sums to 1 across classes

    '''
    shape = logits.size()
    assert len(shape) == 3 # [batch_size, num_sites, num_nodes]

    y_soft = gumbel_softmax_sample(logits, temperature)
    if hard:
        _, k = y_soft.max(1)
        y_hard = torch.zeros(shape, dtype=logits.dtype).scatter_(1, torch.unsqueeze(k, 1), 1.0)

        # This cool bit of code achieves two things:
        # (1) makes the output value exactly one-hot (since we add then subtract y_soft value)
        # (2) makes the gradient equal to y_soft gradient (since we strip all other gradients)
        y = y_hard - y_soft.detach() + y_soft
    else:
        y = y_soft
    return y, y_soft

def full_adj_matrix(U, T, G):
    '''
    All non-zero values of U represent extant clones (leaf nodes of the full tree).
    For each of these non-zero values, we add an edge from parent clone to extant clone.
    '''
    U = U[:,1:] # don't include column for normal cells
    num_leaves = (U > U_CUTOFF).nonzero().shape[0]
    num_internal_nodes = T.shape[0]
    full_adj = torch.nn.functional.pad(input=T, pad=(0, num_leaves, 0, num_leaves), mode='constant', value=0)
    leaf_idx = num_internal_nodes
    # Also add branch lengths (genetic distances) for the edges we're adding.
    # Since all of these edges we're adding represent genetically identical clones,
    # we are going to add a very small but non-zero value.
    full_G = torch.nn.functional.pad(input=G, pad=(0, num_leaves, 0, num_leaves), mode='constant', value=0) if G is not None else None

    # Iterate through the internal nodes that have nonzero values
    for internal_node_idx in (U > U_CUTOFF).nonzero()[:,1]:
        full_adj[internal_node_idx, leaf_idx] = 1
        if G is not None:
            full_G[internal_node_idx, leaf_idx] = G_IDENTICAL_CLONE_VALUE
        leaf_idx += 1
    return full_adj, full_G, num_leaves
    
def compute_u_loss(psi, ref, var, B, weights):
    '''
    Computes loss for latent variable U (dim: num_internal_nodes x num_sites)
    '''
    # Using the softmax enforces that the row sums are 1, since the proprtions of
    # subclones in a given site should sum to 1
    U = torch.softmax(psi, dim=1)
    batch_size = U.shape[0]
    omega_v = no_cna_omega(ref.shape)
    u_loss, loss_dict_b = subclonal_presence_objective(ref, var, omega_v, U, B, weights)
    return U, u_loss

def get_full_G(T, G, poly_res, i=-1, max_iter=-1):
    '''
    If resolving polytomies, dynamically calculate branch lengths between
    polytomy nodes and the "resolver" nodes as the average of the branch 
    lengths of the resolver nodes' children
    '''
    if G == None:
        return None

    if poly_res == None:
        return vutil.repeat_n(G, T.shape[0])

    full_G = vutil.repeat_n(G, T.shape[0])
    # !!TODO: What should the genetic distance be??
    # resolver_indices = poly_res.resolver_indices
    # for batch_idx in range(T.shape[0]):
    #     for res_idx in resolver_indices:
    #         parent_idx = poly_res.resolver_index_to_parent_idx[res_idx]
    #         res_children = vutil.get_child_indices(T[batch_idx], [res_idx])
    #         # no children for this resolver node, so keep the original branch length
    #         if len(res_children) == 0:
    #             avg = G[parent_idx, res_idx]
    #         else:
    #             avg = torch.mean(G[parent_idx, res_children])
    #         full_G[batch_idx, parent_idx, res_idx] = avg
    # LAST_G = full_G
    return full_G

def compute_v_loss(U, X, T, ref, var, p, G, O, v_temp, t_temp, hard, weights, i, max_iter, poly_res=None):
    '''
    Computes loss for X (dim: batch_size x num_internal_nodes x num_sites)
    '''

    def stack_vertex_labeling(U, X, p, T):
        '''
        Use U and X (both of size batch_size x num_internal_nodes x num_sites)
        to get the anatomical sites of the leaf nodes and the internal nodes (respectively). 
        Stack the root labeling to get the full vertex labeling V. 
        '''
        U = U[:,1:] # don't include column for normal cells
        num_sites = U.shape[0]
        L = torch.nn.functional.one_hot((U > U_CUTOFF).nonzero()[:,0], num_classes=num_sites).T
        # Expand leaf node labeling L to be repeated batch_size times
        bs = X.shape[0]
        L = vutil.repeat_n(L, bs)
        
        if poly_res != None:
            # order is: given internal nodes, new poly nodes, leaf nodes from U)
            full_vert_labeling = torch.cat((X, vutil.repeat_n(poly_res.resolver_labeling, bs), L), dim=2)
        else:
            full_vert_labeling = torch.cat((X, L), dim=2)

        p = vutil.repeat_n(p, bs)
        # Concatenate the left part, new column, and right part along the second dimension
        return torch.cat((p, full_vert_labeling), dim=2)

    softmax_X, softmax_X_soft = gumbel_softmax(X, v_temp, hard)
    V = stack_vertex_labeling(U, softmax_X, p, T)

    bs = X.shape[0]
    if poly_res != None:
        softmax_pol_res, _ = gumbel_softmax(poly_res.latent_var, t_temp, hard)
        T = vutil.repeat_n(T, bs)
        T[:,:,poly_res.children_of_polys] = softmax_pol_res
    else:
        T = vutil.repeat_n(T, bs)

    if G != None:
        G = get_full_G(T, G, poly_res, i, max_iter)
    loss, loss_components = ancestral_labeling_objective(V, softmax_X_soft, T, G, O, p, weights, i, max_iter)
    return V, loss, loss_components, softmax_X_soft, T
    
def get_avg_loss_components(loss_components):
    '''
    Calculate the averages of each loss component (e.g. "nll" 
    (negative log likelihood), "mig" (migration number), etc.)
    '''
    d = {}
    for key in loss_components:
        if isinstance(loss_components[key], float) or isinstance(loss_components[key], int):
            d[key] = loss_components[key]
        else:
            d[key] = torch.mean(loss_components[key])
    return d

class VertexLabelingSolution:
    def __init__(self, loss, V, soft_V, T, G, node_idx_to_label, mig_weight, comig_weight, seed_weight, i):
        self.loss = loss
        self.V = V
        self.soft_V = soft_V
        self.T = T
        self.G = G
        self.node_idx_to_label = node_idx_to_label
        self.mig_weight = mig_weight
        self.seed_weight = seed_weight
        self.comig_weight = comig_weight
        self.i = i

   # override the comparison operator
    def __lt__(self, other):
        return self.loss < other.loss


def get_weights_with_best_eval_loss(solutions, weights, O, p):
    '''
    Which parsimony weights give the best gen dist/organotrop
    '''    
    weights_to_loss = {}
    
    out = []
    for i, soln in enumerate(solutions):
        V, soft_V, T, G, mig_weight, seed_weight = soln.V, soln.soft_V, soln.T, soln.G, soln.mig_weight, soln.seed_weight
        reshaped_V = add_batch_dim(V)
        reshaped_soft_V = add_batch_dim(soft_V)
        eval_loss, loss_components = ancestral_labeling_objective(reshaped_V, reshaped_soft_V, T, G, O, p, weights)
        weights_at_i = (float(mig_weight), float(seed_weight))
        if weights_at_i not in weights_to_loss:
            weights_to_loss[weights_at_i] = []
        weights_to_loss[weights_at_i].append(float(eval_loss))
        out.append([float(mig_weight), float(seed_weight), float(eval_loss)])

    #print("weights_to_loss", weights_to_loss)
    softmaxed_scores = torch.softmax(torch.tensor([item[2] for item in out]), dim=0)
    weights_to_softmax_loss = dict()
    for score, item in zip(softmaxed_scores, out):
        weights_at_i = (item[0],item[1])
        if weights_at_i not in weights_to_softmax_loss:
            weights_to_softmax_loss[weights_at_i] = []
        weights_to_softmax_loss[weights_at_i].append(score)

    #print("weights_to_softmax_loss", weights_to_softmax_loss)
    weights_to_avg_loss = {k:(sum(weights_to_loss[k])/len(weights_to_loss[k])) for k in weights_to_loss}
    print("weights_to_avg_loss", weights_to_avg_loss)

    best_weights = None
    lowest_loss = min(list(weights_to_avg_loss.values()))

    best_weights = [k for k,v in weights_to_avg_loss.items() if v == lowest_loss][0]
    
    out.insert(0, ["mig weight", "seed weight", "gen dist loss"])
    return best_weights, out

def get_best_solutions(prev_v_losses, v_losses, Vs, soft_Vs, Ts, best_Vs, best_soft_Vs, best_Ts):
    '''
    Only update solutions where the loss has decreased
    '''
    # First iteration
    if best_Vs == None:
        return Vs, soft_Vs, Ts
    indices_to_update = torch.nonzero(v_losses < prev_v_losses).squeeze()
    best_Vs[indices_to_update] = Vs[indices_to_update]
    best_soft_Vs[indices_to_update] = soft_Vs[indices_to_update]
    best_Ts[indices_to_update] = Ts[indices_to_update]
    return best_Vs, best_soft_Vs, best_Ts

from collections import OrderedDict
import math

class PolytomyResolver():

    def __init__(self, input_T, T, G, U, num_leaves, bs, node_idx_to_label, nodes_w_polys, resolver_sites):
        '''
        This is post U matrix estimation, so T already has leaf nodes.
        '''
        
        # 1. nodes_w_polys are the nodes have polytomies
        #print("nodes_w_polys", nodes_w_polys, "resolver_sites", resolver_sites)
        # 2. Pad the adjacency matrix so that there's room for the new resolver nodes
        # (we place them in this order: given internal nodes, new resolver nodes, leaf nodes from U)
        num_new_nodes = 0
        for r in resolver_sites:
            num_new_nodes += len(r)
       # print("num_new_nodes", num_new_nodes)
        num_internal_nodes = T.shape[0]-num_leaves
        T = torch.nn.functional.pad(T, pad=(0, num_new_nodes, 0, num_new_nodes), mode='constant', value=0)
        # 3. Shift T and G to make room for the new indices (so the order is input internal nodes, new poly nodes, leaves)
        idx1 = num_internal_nodes
        idx2 = num_internal_nodes+num_leaves
        T = torch.cat((T[:,:idx1], T[:,idx2:], T[:,idx1:idx2]), dim=1)
        if G != None:
            G = torch.nn.functional.pad(G, pad=(0, num_new_nodes, 0, num_new_nodes), mode='constant', value=0)
            G = torch.cat((G[:,:idx1], G[:,idx2:], G[:,idx1:idx2]), dim=1)

        # 3. Get each polytomy's children (these are the positions we have to relearn)
        children_of_polys = vutil.get_child_indices(T, nodes_w_polys)
        #print("children_of_polys", children_of_polys)

        # 4. Initialize a matrix to learn the polytomy structure
        num_nodes_full_tree = T.shape[0]
        poly_adj_matrix = vutil.repeat_n(torch.zeros((num_nodes_full_tree, len(children_of_polys)), dtype=torch.float32), bs)
        resolver_indices = [x for x in range(num_internal_nodes, num_internal_nodes+num_new_nodes)]
        #print("resolver_indices", resolver_indices)

        nodes_w_polys_to_resolver_indices = OrderedDict()
        start_new_node_idx = resolver_indices[0]
        for parent_idx, r in zip(nodes_w_polys, resolver_sites):
            num_new_nodes_for_poly = len(r)
            if parent_idx not in nodes_w_polys_to_resolver_indices:
                nodes_w_polys_to_resolver_indices[parent_idx] = []

            for i in range(start_new_node_idx, start_new_node_idx+num_new_nodes_for_poly):
                nodes_w_polys_to_resolver_indices[parent_idx].append(i)
            start_new_node_idx += num_new_nodes_for_poly
        #print("nodes_w_polys_to_resolver_indices", nodes_w_polys_to_resolver_indices)

        resolver_labeling = torch.zeros(U.shape[0], len(resolver_indices))
        t = 0
        for sites in resolver_sites:
            for site in sites:
                resolver_labeling[site, t] = 1
                t += 1
        #print("resolver_labeling", resolver_labeling)

        # print("resolver_indices", resolver_indices)
        offset = 0
        for parent_idx in nodes_w_polys:
            child_indices = vutil.get_child_indices(T, [parent_idx])
            # make the children of polytomies start out as children of their og parent
            # with the option to "switch" to being the child of the new poly node
            poly_adj_matrix[:,parent_idx,offset:(offset+len(child_indices))] = 1.0
            # we only want to let these children choose between being the child
            # of their original parent or the child of this new poly node, which
            # we can do by setting all other indices to -inf
            mask = torch.ones(num_nodes_full_tree, dtype=torch.bool)
            new_nodes = nodes_w_polys_to_resolver_indices[parent_idx]
            mask_indices = new_nodes + [parent_idx]
            #print("parent_idx", parent_idx, "mask_indices", mask_indices)
            mask[[mask_indices]] = 0
            poly_adj_matrix[:,mask,offset:(offset+len(child_indices))] = float("-inf")
            offset += len(child_indices)

        poly_adj_matrix.requires_grad = True
        
        # 5. Initialize potential new nodes as children of the polytomy nodes
        for i in nodes_w_polys:
            for j in nodes_w_polys_to_resolver_indices[i]:
                T[i,j] = 1.0
                node_idx_to_label[j] = f"{i}pol{j}"
                if G != None:
                    G[i,j] = G_IDENTICAL_CLONE_VALUE

        # 6. The genetic distance between a new node and its potential
        # new children which "switch" is the same distance between the new
        # node's parent and the child
        resolver_index_to_parent_idx = {}
        for poly_node in nodes_w_polys_to_resolver_indices:
            new_nodes = nodes_w_polys_to_resolver_indices[poly_node]
            for new_node_idx in new_nodes:
                resolver_index_to_parent_idx[new_node_idx] = poly_node
        #print("resolver_index_to_parent_idx", resolver_index_to_parent_idx)

        if G != None:
            for new_node_idx in resolver_indices:
                parent_idx = resolver_index_to_parent_idx[new_node_idx]
                potential_child_indices = vutil.get_child_indices(T, [parent_idx])
                for child_idx in potential_child_indices:
                    G[new_node_idx, child_idx] = G[parent_idx, child_idx]

        self.latent_var = poly_adj_matrix
        self.nodes_w_polys = nodes_w_polys
        self.children_of_polys = children_of_polys
        self.resolver_indices = resolver_indices
        self.T = T
        self.G = G
        self.node_idx_to_label = node_idx_to_label
        self.resolver_index_to_parent_idx = resolver_index_to_parent_idx
        self.resolver_labeling = resolver_labeling

def is_same_mig_hist_with_node_removed(T, V, remove_idx, poly_res, p):
    '''
    Returns True if migration #, comigration # and seeding # are
    the same after removing node at index remove_idx
    '''
    prev_m, prev_c, prev_s, _, _ = get_ancestral_labeling_metrics(add_batch_dim(V), T, None, None, p)
    # Attach all the children of the candidate removal node to
    # its parent, and then check if that changes the migration history or not

    candidate_T = T.clone().detach()
    candidate_V = V.clone().detach()
    parent_idx = np.where(T[:,remove_idx] > 0)[0][0]
    child_indices = vutil.get_child_indices(T, [remove_idx])
    for child_idx in child_indices:
        candidate_T[parent_idx,child_idx] = 1.0
    candidate_T = np.delete(candidate_T, remove_idx, 0)
    candidate_T = np.delete(candidate_T, remove_idx, 1)
    candidate_V = np.delete(candidate_V, remove_idx, 1)
    new_m, new_c, new_s, _, _ = get_ancestral_labeling_metrics(add_batch_dim(candidate_V), candidate_T, None, None, p)
    
    return ((prev_m == new_m) and (prev_c == new_c) and (prev_s == new_s))

def remove_nodes(removal_indices, V, T, G, node_idx_to_label):
    '''
    '''
    # TODO: test if we need this, pytorch acting weird about no grad mode
    T = T.clone().detach()
    V = V.clone().detach()
    
    # Attach children of the node to remove to their original parent
    for remove_idx in removal_indices:
        parent_idx = np.where(T[:,remove_idx] > 0)[0][0]
        child_indices = vutil.get_child_indices(T, [remove_idx])
        for child_idx in child_indices:
            T[parent_idx,child_idx] = 1.0
    # Remove indices from T, V, soft V and G
    T = np.delete(T, removal_indices, 0)
    T = np.delete(T, removal_indices, 1)
    V = np.delete(V, removal_indices, 1)
    if G != None: 
        G = G.clone().detach()
        G = np.delete(G, removal_indices, 0)
        G = np.delete(G, removal_indices, 1)

    # Reindex the idx to label dict
    copy_node_idx_to_label = copy.deepcopy(node_idx_to_label)
    for idx in removal_indices:
        del copy_node_idx_to_label[idx]
    new_node_idx_to_label = dict()
    for i,key in enumerate(sorted(list(copy_node_idx_to_label.keys()))):
        new_node_idx_to_label[i] = copy_node_idx_to_label[key]
    return V, T, G, new_node_idx_to_label

def remove_extra_resolver_nodes(best_Vs, Ts, node_idx_to_label, G, poly_res, p):
    '''
    If there are any resolver nodes that were added to resolve polytomies but they 
    weren't used (i.e. 1. they have no children or 2. they don't change the 
    migration history), remove them
    '''

    if poly_res == None:
        return best_Vs, Ts, [G for _ in range(len(best_Vs))], [node_idx_to_label for _ in range(len(best_Vs))]

    out_Vs, out_Ts, out_Gs, out_node_idx_to_labels = [],[],[],[]
    for V, T in zip(best_Vs, Ts):
        nodes_to_remove = []
        for new_node_idx in poly_res.resolver_indices:
            children_of_new_node = vutil.get_child_indices(T, [new_node_idx])
            if len(children_of_new_node) == 0:
                nodes_to_remove.append(new_node_idx)
            elif is_same_mig_hist_with_node_removed(T, V, new_node_idx, poly_res, p):
                nodes_to_remove.append(new_node_idx)

        # Genetic distance matrix must get dynamically updated based on the resolved polytomies
        if G != None:
            resolved_G = get_full_G(add_batch_dim(T), G, poly_res)[0]
        else:
            resolved_G = None
        new_V, new_T, new_G, new_node_idx_to_label = remove_nodes(nodes_to_remove, V, T, resolved_G, node_idx_to_label)
        out_Vs.append(new_V)
        out_Ts.append(new_T)
        out_Gs.append(new_G)
        out_node_idx_to_labels.append(new_node_idx_to_label)
    return out_Vs, out_Ts, out_Gs, out_node_idx_to_labels


# def is_less_parsimonious(pars_metrics, best_pars_metrics):
    
#     for best_pars_metric in best_pars_metrics:
#         if not (pars_metrics[0] <= best_pars_metric[0] and pars_metrics[1] <= best_pars_metric[1] and pars_metrics[2] <= best_pars_metric[2]):
#             return True
#     return False

def made_mistake(solution, U):
    V = solution.V
    A = solution.T
    VA = V @ A
    W = VA.T @ VA # 1 if two nodes' parents are the same color
    X = V.T @ V # 1 if two nodes are the same color
    Y = torch.sum(torch.mul(VA.T, 1-V.T), axis=1) # Y has a 1 for every node where its parent has a diff color
    nonzero_indices = torch.nonzero(Y).squeeze()
    U = U[:,1:] # don't include column for normal cells
    num_sites = U.shape[0]
    L = torch.nn.functional.one_hot((U > U_CUTOFF).nonzero()[:,0], num_classes=num_sites).T
    num_internal_nodes = A.shape[0] - L.shape[1]
    if nonzero_indices.dim() == 0:
        return False
    # print(nonzero_indices)
    for mig_node in nonzero_indices:
        # it's a leaf node itself!
        if mig_node > (num_internal_nodes-1):
            continue
        if not vutil.has_leaf_node(A, int(mig_node), num_internal_nodes):
            return True
    return False
    

def prune_bad_histories(solutions, O, p, weights, mode, U):
    '''
    Detect mistakes and prune out those histories (this is specifically
    in datasets where there are branches of the tree with no U leaf nodes)
    Removes any solutions that are > parsimony_eps parsimony metrics
    away from the best solution
    '''
    parsimony_eps = 1

    def keep_tree(cand_metric, best_sum, best_pars_metrics):
        if abs(best_sum - sum(cand_metric)) > parsimony_eps:
            return False
        if sum(cand_metric) == best_sum:
            return True
        # All these trees are +/- parsimony_eps, and they need to 
        # improve at least one of the metrics compared to the best_pars_metrics
        for best_metric in best_pars_metrics:
            if cand_metric[0] < best_metric[0] or cand_metric[1] < best_metric[1] or cand_metric[2] < best_metric[2]:
                return True
        return False

    # Collect each solution's parsimony metrics
    all_pars_metrics = []
    for soln in solutions:
        V, T, G = soln.V, soln.T, soln.G
        m, c, s, _, _ = get_ancestral_labeling_metrics(add_batch_dim(V), T, G, O, p)
        all_pars_metrics.append((int(m), int(c), int(s)))

    # Find the best parsimony sum
    best_sum = float("inf")
    for i, pars_metrics in enumerate(all_pars_metrics):
        pars_sum = sum(pars_metrics)
        if pars_sum < best_sum:
            best_sum = pars_sum

    # Find all pars metrics combinations that match the best sum
    best_pars_metrics = set()
    for i, pars_metrics in enumerate(all_pars_metrics):
        pars_sum = sum(pars_metrics)
        if pars_sum == best_sum:
            best_pars_metrics.add(pars_metrics)

    pruned_solutions = []
    
    # Go through and prune any solutions that are worse than the best sum
    # or made any mistakes
    for soln, pars_metrics in zip(solutions, all_pars_metrics):
        if keep_tree(pars_metrics, best_sum, best_pars_metrics) and not made_mistake(soln, U):
            pruned_solutions.append(soln)

    if len(pruned_solutions) == 0: 
        print("no solutions without mistakes detected")
        # ideally this doesn't happen, but remove mistake detection so 
        # that we return some results
        for soln, pars_metrics in zip(solutions, all_pars_metrics):
            if keep_tree(pars_metrics, best_sum, best_pars_metrics):
                pruned_solutions.append(soln)

    return pruned_solutions

def _get_best_final_solutions(mode, best_Vs, best_soft_Vs, Ts, Gs, O, p, weights, print_config, 
                              poly_res, idx_label_dicts, output_dir, run_name, U, needs_pruning=True):
    '''
    Weigh solutions differently based on evaluate or calibrate mode. 
    Return the top k solutions
    '''

    # 1. Initialize weights for loss calculations based on what mode we're in
    input_weights = weights
    if mode == "evaluate":
        weights = weights
    elif mode == 'calibrate': 
        # Evaluate most parsimonious trees via gen. dist. and organotropism
        weights = met.Weights(mig=[0.0], comig=0.0, seed_site=[0.0], gen_dist=weights.gen_dist, organotrop=weights.organotrop, data_fit=0.0, reg=0.0, entropy=0.0)

    # 2. Calculate loss for each solution
    bs = len(best_Vs)
    mig_weights = get_mig_weight_vector(bs, input_weights)
    seed_weights = get_seed_site_weight_vector(bs, input_weights)
    solutions = []
    for i, (V, soft_V, T, G, idx_label_dict) in enumerate(zip(best_Vs, best_soft_Vs, Ts, Gs, idx_label_dicts)):
        reshaped_V = add_batch_dim(V)
        reshaped_soft_V = add_batch_dim(soft_V)        
        loss, loss_components = ancestral_labeling_objective(reshaped_V, reshaped_soft_V, T, G, O, p, weights)
        # TODO: take mig_delta as input?
        solutions.append(VertexLabelingSolution(loss, V, soft_V, T, G, idx_label_dict, mig_weights[i], input_weights.comig, seed_weights[i], i))
    
    # 3. Prune bad histories (anything less parsimonious than the best solution we find)
    if needs_pruning:
        pruned_solutions = prune_bad_histories(solutions, O, p, input_weights, mode, U)
    else:
        pruned_solutions = solutions

    # 4. Make the solutions unique
    unique_solutions = set()
    final_solutions = []
    seeding_outputs = [["Evaluation loss", "Seeding pattern"]]
    for soln in pruned_solutions:
        tree = vutil.LabeledTree(soln.T, soln.V)
        if tree not in unique_solutions:
            final_solutions.append(soln)
            unique_solutions.add(tree)
        seeding_outputs.append([float(soln.loss), plot_util.get_seeding_pattern(soln.V, soln.T)])

    # 5. Sort the solutions from lowest to highest loss
    final_solutions = sorted(list(final_solutions))

    # 6. Return the k best solutions
    k = print_config.k_best_trees if len(final_solutions) >= print_config.k_best_trees else len(final_solutions)
    final_solutions = final_solutions[:k]
    return final_solutions, seeding_outputs

def get_best_final_solutions(mode, best_Vs, best_soft_Vs, Ts, G, O, p, weights, print_config, 
                             poly_res, node_idx_to_label, output_dir, run_name, U):
    '''
    Prune unecessary poly nodes (if they weren't used), then weigh solutions  
    differently based on evaluate or calibrate mode. 
    Return the top k solutions
    '''
    # 1. Remove any extra resolver nodes that don't actually help
    best_Vs, Ts, Gs, idx_label_dicts = remove_extra_resolver_nodes(best_Vs, Ts, node_idx_to_label, G, poly_res, p)

    return _get_best_final_solutions(mode, best_Vs, best_soft_Vs, Ts, Gs, O, p, weights, print_config, 
                                     poly_res, idx_label_dicts, output_dir, run_name, U)

def init_X(batch_size, num_sites, num_nodes_to_label, weight_init_primary, p, input_T, T, U):
    # We're learning X, which is the vertex labeling of the internal nodes
    X = torch.rand(batch_size, num_sites, num_nodes_to_label)
    #print("X", X.shape)
    eta = 2 # TODO: how do we set this parameter?
    if weight_init_primary:
        # for each node, find which sites to bias labeling towards using
        # the sites it and its children are detected in
        nodes_w_children, sites = vutil.get_k_or_more_children_nodes(input_T, T, U, 1, 1, cutoff=False)
        #print("nodes_w_children", nodes_w_children, "sites", sites)
        prim_site_idx = torch.nonzero(p)[0][0]
        #print("p", p, "prim_site_idx", prim_site_idx)
        X[:,prim_site_idx,:] = eta

        for node_idx, sites in zip(nodes_w_children, sites):
            if node_idx == 0:
                continue # we know the root labeling
            idx = node_idx - 1
            for site_idx in sites:
                X[:,site_idx,idx] = eta
        #print("X", X[0])
    X.requires_grad = True
    return X

def update_adj_matrix(itr, max_iter):
    if itr == -1:
        return True
    global SOLVE_POLYS
    if not SOLVE_POLYS:
        return False
    return itr > max_iter*1/3 and itr < max_iter*2/3
    # return itr > max_iter*1/2

def remove_leaf_nodes_idx_to_label_dicts(dicts):
    '''
    After running migration history inference, the leaf nodes 
    (U nodes) get added to the idx to label dicts when plotting 
    and saving to pickle files, so we don't want to do that twice
    when in calibrate mode
    '''
    new_dicts = []
    for i,dct in enumerate(dicts):
        new_dicts.append(copy.deepcopy(dct))
        for key in dct:
            if dct[key][1] == True:
                del new_dicts[i][key]
            else:
                new_dicts[i][key] = new_dicts[i][key][0]
    return new_dicts

# TODO: should we not take weights as input for this?
def calibrate(Ts, ref_matrices, var_matrices, ordered_sites, primary_sites, 
              node_idx_to_labels, weights, print_config, output_dir, run_names,
              Gs=None, Os=None, max_iter=100, lr=0.1, init_temp=40, final_temp=0.01,
              batch_size=64, custom_colors=None, weight_init_primary=True,  solve_polytomies=False):
    if not (len(Ts) == len(ref_matrices) == len(var_matrices) == len(ordered_sites) == len(primary_sites) == len(node_idx_to_labels) == len(run_names)):
        raise ValueError("Inputs ref_matrices, var_matrices, ordered_sites, primary_sites, node_idx_to_labels and run_names must have equal length (length = to number of patients in cohort")
    if (Gs == None and Os == None):
        raise ValueError("In calibrate mode, either Gs or Os should be input")
    if (Gs != None and len(Gs) != len(Ts)):
        raise ValueError("Length of Ts and Gs must be equal")
    if (Os != None and len(Os) != len(Ts)):
        raise ValueError("Length of Ts and Os must be equal")


    def _convert_list_of_numpys_to_tensors(lst):
        return [torch.tensor(x) for x in lst]

    input_weights = copy.deepcopy(weights)
    organotrop_weight = 0.0 if Os == None else (weights.organotrop if weights != None else 0.1)
    gen_dist_weight = 0.0 if Gs == None else (weights.gen_dist if weights != None else 0.1)

    weights = met.Weights(mig=DEFAULT_CALIBRATE_MIG_WEIGHTS, comig=DEFAULT_CALIBRATE_COMIG_WEIGHTS, 
                          seed_site=DEFAULT_CALIBRATE_SEED_WEIGHTS, gen_dist=gen_dist_weight, 
                          organotrop=organotrop_weight)
    # Don't spend time making visualizations for calibrated trees
    visualize = print_config.visualize
    input_k = print_config.k_best_trees
    print_config.visualize = False
    print_config.k_best_trees = batch_size

    calibrate_dir = os.path.join(output_dir, "calibrate")

    print(f"Saving results to {calibrate_dir}")

    if os.path.exists(calibrate_dir):
        shutil.rmtree(calibrate_dir)
        print(f"Overwriting existing directory at {calibrate_dir}")
    
    os.makedirs(calibrate_dir)

    # 1. Go through each patient and get migration history in calibrate mode
    for i in range(len(Ts)):
        print("Calibrating for patient:", run_names[i])
        G = Gs[i] if Gs != None else None
        O = Os[i] if Os != None else None
        get_migration_history(Ts[i], ref_matrices[i], var_matrices[i], ordered_sites[i], primary_sites[i],
                              node_idx_to_labels[i], weights, print_config, calibrate_dir, f"{run_names[i]}", 
                              G=G, O=O, max_iter=max_iter, lr=lr, init_temp=init_temp, final_temp=final_temp,
                              batch_size=batch_size, custom_colors=custom_colors, weight_init_primary=weight_init_primary, 
                              mode="calibrate", solve_polytomies=solve_polytomies)

    # 2. Find the best theta for this cohort
    best_theta = eutil.get_max_cross_ent_thetas(pickle_file_dirs=[calibrate_dir])
    rounded_best_theta = [round(v,3) for v in best_theta]
    with open(os.path.join(calibrate_dir, "best_theta.json"), 'w') as json_file:
        json.dump(rounded_best_theta, json_file, indent=2)
        
    # 3. Recalibrate trees using the best thetas
    print_config.visualize = visualize
    print_config.k_best_trees = input_k
    weights = met.Weights(mig=[best_theta[0]], comig=best_theta[1], seed_site=[best_theta[2]],
                          gen_dist=input_weights.gen_dist if Gs != None else 0.0, organotrop=input_weights.organotrop if Os != None else 0.0)
    
    for i in range(len(Ts)):
        O = Os[i] if Os != None else None
        with gzip.open(os.path.join(calibrate_dir, f"{run_names[i]}.pkl.gz"), 'rb') as f:
            pckl = pickle.load(f)
        saved_Ts = _convert_list_of_numpys_to_tensors(pckl[OUT_ADJ_KEY])
        saved_Vs = _convert_list_of_numpys_to_tensors(pckl[OUT_LABElING_KEY])
        saved_soft_Vs = _convert_list_of_numpys_to_tensors(pckl[OUT_SOFTV_KEY])
        saved_U = torch.tensor(pckl[OUT_SUB_PRES_KEY])
        saved_Gs = _convert_list_of_numpys_to_tensors(pckl[OUT_GEN_DIST_KEY])
        saved_idx_to_label_dicts = remove_leaf_nodes_idx_to_label_dicts(pckl[OUT_IDX_LABEL_KEY])
        primary_idx = ordered_sites[i].index(primary_sites[i])
        p = torch.nn.functional.one_hot(torch.tensor([primary_idx]), num_classes=len(ordered_sites[i])).T
        
        final_solutions, _ = _get_best_final_solutions('evaluate', saved_Vs, saved_soft_Vs, saved_Ts, saved_Gs, O, p, weights,
                                                       print_config, None, saved_idx_to_label_dicts, calibrate_dir, run_names[i], 
                                                       saved_U, needs_pruning=False)
        
        plot_util.print_best_trees(final_solutions, saved_U, ref_matrices[i], var_matrices[i], O,
                                  weights, ordered_sites[i], print_config, 
                                  custom_colors, primary_sites[i], 
                                  calibrate_dir, run_names[i])

    return best_theta


def get_migration_history(T, ref, var, ordered_sites, primary_site, node_idx_to_label,
                          weights, print_config, output_dir, run_name, 
                          G=None, O=None, max_iter=100, lr=0.1, init_temp=40, final_temp=0.01,
                          batch_size=64, custom_colors=None, weight_init_primary=True, mode="evaluate",
                          solve_polytomies=False):
    '''
    Args:
        T: numpy ndarray or torch tensor (shape: num_internal_nodes x num_internal_nodes). Adjacency matrix (directed) of the internal nodes.
        
        ref: numpy ndarray or torch tensor (shape: num_anatomical_sites x num_mutation_clusters). Reference matrix, i.e., num. reads that map to reference allele
        
        var: numpy ndarray or torch tensor (shape:  num_anatomical_sites x num_mutation_clusters). Variant matrix, i.e., num. reads that map to variant allele
        
        ordered_sites: list of the anatomical site names (e.g. ["breast", "lung_met"]) with length =  num_anatomical_sites) and 
        the order matches the order of sites in the ref and var
        
        primary_site: name of the primary site (must be an element of ordered_sites)

        node_idx_to_label: dictionary mapping vertex indices (corresponding to their index in T) to custom labels
        for plotting

        weights: Weight object for how much to penalize each component of the loss
        
        print_config: PrintConfig object with options on how to visualize output
        
        output_dir: path for where to save output trees to

        run_name: e.g. patient name, used for naming output files.

    Optional:
        G: numpy ndarray or torch tensor (shape: num_internal_nodes x num_internal_nodes).
        Matrix of genetic distances between internal nodes.
        Lower values indicate lower branch lengths, i.e. more genetically similar.
        
        O: numpy ndarray or torch tensor (shape: num_anatomical_sites x  num_anatomical_sites).
        Matrix of organotropism values between sites.

        weight_init_primary: whether to initialize weights higher to favor vertex labeling of primary for all internal nodes

        mode: can be "evaluate" or "calibrate"

    Returns:
        # TODO: return info for k best trees
        Corresponding info on the *best* tree:
        (1) edges of the tree (e.g. [('0', '1'), ('1', '2;3')])
        (2) vertex labeling as a dictionary (e.g. {'0': 'P', '1;3': 'M1'}),
        (3) edges for the migration graph (e.g. [('P', 'M1')])
        (4) dictionary w/ loss values for each component of the loss
        (5) how long (in seconds) the algorithm took to run
    '''
    if not (T.shape[0] == T.shape[1]):
        raise ValueError(f"Number of tree nodes should be consistent (T.shape[0] == T.shape[1])")
    if (T.shape[0] != len(node_idx_to_label)):
        raise ValueError(f"Number of node_idx_to_label needs to equal shape of adjacency matrix.")
    if ref.shape != var.shape:
        raise ValueError(f"ref and var must have identical shape, got {ref.shape} and {var.shape}")
    if not (ref.shape[1] == var.shape[1] == T.shape[0]):
        raise ValueError(f"Number of mutations/mutation clusters should be consistent (ref.shape[1] == var.shape[1] == T.shape[0])")
    if not (ref.shape[0] == var.shape[0] == len(ordered_sites)):   
        raise ValueError(f"Length of ordered_sites should be equal to ref and var dim 0")
    if not vutil.is_tree(T):
        raise ValueError("Adjacency matrix T is empty or not a tree.")
    if not primary_site in ordered_sites:
        raise ValueError(f"{primary_site} not in ordered_sites: {ordered_sites}")
    if (weights.gen_dist == 0.0 and G != None):
        raise ValueError(f"G matrix was given but genetic distance parameter of weights is 0. Please set genetic distance weight to > 0.")
    if (weights.gen_dist != 0.0 and G == None):
        raise ValueError(f"G matrix was not given but genetic distance parameter of weights is non-zero. Please pass a G matrix.")
    if (weights.organotrop == 0.0 and O != None):
        raise ValueError(f"O matrix was given but organotropism parameter of weights is 0. Please set organotropism weight to > 0.")
    if (weights.organotrop != 0.0 and O == None):
        raise ValueError(f"O matrix was not given but organotropism parameter of weights is non-zero. Please pass an O matrix.")
    if mode != 'calibrate' and mode != 'evaluate':
        raise ValueError(f"Valid modes are 'evaluate' and 'calibrate'")
    if mode == 'calibrate':
        if not ((weights.organotrop != 0.0 and O != None) or (weights.gen_dist != 0.0 and G != None)):
            raise ValueError(f"In calibrate mode, either organotropism or genetic distance matrices and weights must be set.")
    for label in list(node_idx_to_label.values()):
        if ":" in label:
            raise ValueError(f"Unfortunately our visualization code uses pydot, which does not allow colons (:) in node names. Please use a different separator in the values of node_idx_to_label")

    # Don't alter the inputs in place
    T = copy.deepcopy(T)
    G = copy.deepcopy(G)
    ref = copy.deepcopy(ref)
    var = copy.deepcopy(var)
    ordered_sites = copy.deepcopy(ordered_sites)
    primary_site = copy.deepcopy(primary_site)
    node_idx_to_label = copy.deepcopy(node_idx_to_label)

    root_idx = vutil.get_root_index(T)
    if root_idx != 0:
        print(f"Restructuring adjacency matrix for {run_name} since root node is not at index 0")
        T, ref, var, node_idx_to_label, G = vutil.restructure_matrices(T, ref, var, node_idx_to_label, G)
    assert(vutil.get_root_index(T) == 0)

    # TODO: convert to torch float 32 if its not already
    if not torch.is_tensor(T):
        T = torch.tensor(T, dtype=torch.float32)
    if not torch.is_tensor(ref):
        ref = torch.tensor(ref, dtype=torch.float32)
    if not torch.is_tensor(var):
        var = torch.tensor(var, dtype=torch.float32)

    primary_idx = ordered_sites.index(primary_site)
    p = torch.nn.functional.one_hot(torch.tensor([primary_idx]), num_classes=len(ordered_sites)).T
    num_sites = ref.shape[0]
    num_internal_nodes = T.shape[0]

    # If we don't know the anatomical site of the primary tumor, we need to learn it
    num_nodes_to_label = -1
   
    assert(p.shape[1] == 1)
    assert(p.shape[0] == ref.shape[0]) # num_anatomical_sites
    num_nodes_to_label = num_internal_nodes - 1 # we don't need to learn the root labeling
    prim_site_idx = torch.nonzero(p)[0][0]
    primary_site_label = ordered_sites[prim_site_idx]

    # We're learning psi, which is the mixture matrix U (U = softmax(psi)), and tells us the existence
    # and anatomical locations of the extant clones (U > U_CUTOFF)
    psi = -1 * torch.rand(num_sites, num_internal_nodes + 1) # an extra column for normal cells
    psi.requires_grad = True 
    u_optimizer = torch.optim.Adam([psi], lr=lr)

    B = vutil.get_mutation_matrix_tensor(T)
    # add a row of zeros to account for the non-cancerous root node
    B = torch.vstack([torch.zeros(B.shape[1]), B])
    # add a column of ones to indicate that every subclone has the non-cancerous mutations
    B = torch.hstack([torch.ones(B.shape[0]).reshape(-1,1), B])

    # reinitialize globals as None (if module is not reloaded in between runs, this can cause problems)
    global RANDOM_VALS, LAST_P, G_IDENTICAL_CLONE_VALUE, SOLVE_POLYS
    RANDOM_VALS = None
    LAST_P = None
    SOLVE_POLYS = solve_polytomies

    if G != None:
        G_IDENTICAL_CLONE_VALUE = torch.min(G[(G != 0)])/2.0

    start_time = datetime.datetime.now()

    input_T = copy.deepcopy(T)
    i = 0
    u_prev = psi
    u_diff = 1e9
    ############ Step 1, find a MAP estimate of U ############
    while u_diff > 1e-4 and i < 200:
        u_optimizer.zero_grad()
        U, u_loss = compute_u_loss(psi, ref, var, B, weights)
        u_loss.backward()
        u_optimizer.step()
        u_diff = torch.abs(torch.norm(u_prev - U))
        u_prev = U
        i += 1
    
    T, G, num_leaves = full_adj_matrix(U, T, G)
    if solve_polytomies:
        nodes_w_polys, resolver_sites = vutil.get_k_or_more_children_nodes(input_T, T, U, 3, 2)
        if len(nodes_w_polys) == 0:
            print("No potential polytomies to solve, not resolving polytomies.")
            poly_res = None
            solve_polytomies = False
        else:
            poly_res = PolytomyResolver(input_T, T, G, U, num_leaves, batch_size, node_idx_to_label, nodes_w_polys, resolver_sites)
            T = poly_res.T
            print("T", T.shape)
            G = poly_res.G
            node_idx_to_label = poly_res.node_idx_to_label
            #num_nodes_to_label += len(poly_res.resolver_indices)

            poly_optimizer = torch.optim.Adam([poly_res.latent_var], lr=lr)
            print("num_nodes_to_label", num_nodes_to_label)
    else:
        poly_res = None

    # We're learning X, which is the vertex labeling of the internal nodes
    X = init_X(batch_size, num_sites, num_nodes_to_label, weight_init_primary, p, input_T, T, U)
    
    v_optimizer = torch.optim.Adam([X], lr=lr)
    scheduler = lr_scheduler.LinearLR(v_optimizer, start_factor=1.0, end_factor=0.5, total_iters=max_iter)

    ############ Step 2, sample from V to estimate q(V) ############

    # Temperature and annealing
    v_temp = init_temp
    t_temp = init_temp
    hard = True
    v_anneal_rate = 0.002
    t_anneal_rate = 0.1
    intermediate_data = []
    temps = []
    v_loss_components = []

    # Calculate V only using maximum parsimony metrics
    if mode == 'calibrate':
        # Make a copy of inputs for later
        weights_copy = copy.deepcopy(weights)
        G_copy = copy.deepcopy(G)
        O_copy = copy.deepcopy(O)
        weights = met.Weights(mig=weights.mig, comig=weights.comig, seed_site=weights.seed_site, gen_dist=0.0, organotrop=0.0, data_fit=weights.data_fit, reg=weights.reg, entropy=weights.entropy)
        G, O = None, None
    
    # Keep track of the best trees and losses
    best_Vs, best_soft_Vs, best_Ts, = None, None, None
    prev_v_losses = None
    j = 0
    k = 0
    for i in tqdm(range(max_iter)):

        if solve_polytomies: 
            poly_optimizer.zero_grad()
        v_optimizer.zero_grad()
        Vs, v_losses, loss_comps, soft_Vs, Ts = compute_v_loss(U, X, T, ref, var, p, G, O, v_temp, t_temp, hard, weights, i, max_iter, poly_res)
        mean_loss = torch.sum(v_losses)
        mean_loss.backward()

        if solve_polytomies and update_adj_matrix(i, max_iter):
            poly_optimizer.step()
            if i % 2 == 0:
                t_temp = np.maximum(t_temp * np.exp(-t_anneal_rate * j), final_temp)
            j += 1

        else:
            v_optimizer.step()
            scheduler.step()
            if i % 5 == 0:
                v_temp = np.maximum(v_temp * np.exp(-v_anneal_rate * k), final_temp)
            k += 1

        if print_config.visualize:
            v_loss_components.append(get_avg_loss_components(loss_comps))

        with torch.no_grad():
            best_Vs, best_soft_Vs, best_Ts = get_best_solutions(prev_v_losses, v_losses, Vs, soft_Vs, Ts, best_Vs, best_soft_Vs, best_Ts)
            
        temps.append(v_temp)

        # print("i", i, "v temp", v_temp, "t_temp", t_temp)

        prev_v_losses = v_losses


    time_elapsed = (datetime.datetime.now() - start_time).total_seconds()
    if print_config.verbose:
        print(f"Time elapsed: {time_elapsed}")

    with torch.no_grad():
        if mode == "calibrate":
            # Reload
            G, O = G_copy, O_copy
            weights = weights_copy
        
        final_solutions, seeding_outputs = get_best_final_solutions(mode, best_Vs, best_soft_Vs, best_Ts, G, O, p, weights,
                                                                    print_config, poly_res, node_idx_to_label, output_dir, 
                                                                    run_name, U)

        print("# final solutions:", len(final_solutions))

        if mode == 'calibrate':
            with open(os.path.join(output_dir, f"{run_name}_eval_loss_by_seeding_pattern.txt"), 'w', newline='') as file:
                writer = csv.writer(file)
                writer.writerows(seeding_outputs)


        if print_config.visualize:
            plot_util.plot_loss_components(v_loss_components, weights)
            #plot_util.plot_temps(temps)

        edges, vert_to_site_map, mig_graph_edges, loss_info = plot_util.print_best_trees(final_solutions, U, ref, var, O, 
                                                                                         weights, ordered_sites,print_config, 
                                                                                         custom_colors, primary_site_label, 
                                                                                         output_dir, run_name)

        # avg_tree = plot_util.print_averaged_tree(losses_tensor, V, full_trees, node_idx_to_label, custom_colors,
        #                                     ordered_sites, print_config)

       

    return edges, vert_to_site_map, mig_graph_edges, loss_info, time_elapsed



import numpy as np
import networkx as nx
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import itertools
import pydot
import torch
from networkx.drawing.nx_pydot import to_pydot
from graphviz import Source
from PIL import Image as PILImage
from IPython.display import Image, display
import io
import matplotlib.gridspec as gridspec
import math
import matplotlib.font_manager
from matplotlib import rcParams
import os
import pickle
import pygraphviz as pgv
from collections import deque
import gzip
import string

# TODO: this cyclical import is not great
import metient.lib.vertex_labeling as vert_label
from metient.util.vertex_labeling_util import LabeledTree, get_root_index, tree_iterator
from metient.util.globals import *

import pandas as pd
pd.options.display.float_format = '{:,.3f}'.format

import seaborn as sns

COLORS = ["#6aa84fff","#c27ba0ff", "#e69138ff", "#be5742e1", "#2496c8ff", "#674ea7ff"] + sns.color_palette("Paired").as_hex()


if torch.cuda.is_available():
    torch.set_default_tensor_type(torch.cuda.FloatTensor)

def is_cyclic(G):
    '''
    returns True if graph contains cycles
    '''
    num_nodes = G.size(0)
    visited = torch.zeros(num_nodes, dtype=torch.bool)
    stack = torch.zeros(num_nodes, dtype=torch.bool)

    def dfs(node):
        visited[node] = True
        stack[node] = True

        for neighbor in range(num_nodes):
            if G[node, neighbor] == 1:
                if not visited[neighbor]:
                    if dfs(neighbor):
                        return True
                elif stack[neighbor]:
                    return True

        stack[node] = False
        return False

    for node in range(num_nodes):
        if not visited[node]:
            if dfs(node):
                return True

    return False

def get_seeding_pattern(V, A):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)

    returns: verbal description of the seeding pattern, one of:
    {monoclonal, polyclonal} {single-source, multi-source, reseeding}
    '''
    if not isinstance(V, torch.Tensor):
        V = torch.tensor(V)
    if not isinstance(A, torch.Tensor):
        A = torch.tensor(A)

    G = get_migration_graph(V, A)
    pattern = ""
    # 1) determine if monoclonal (no multi-eges) or polyclonal (multi-edges)
    pattern = "polyclonal " if ((G > 1).any()) else  "monoclonal "

    # 2) determine if single-source seeding (all incoming edges to a site in G 
    # originate from the same site) OR multi-source seeding (at least one site is 
    # seeded from multiple other sites) OR (R) reseeding (at least one site seeds 
    # its originating site)
    non_zero = torch.where(G > 0)
    source_sites = non_zero[0]
    dest_sites = non_zero[1]
    binarized_G = (G != 0).to(torch.int)

    col_sums = torch.sum(binarized_G, axis=0)
    # single-source means that each site is only seeded by ONE other site 
    # (not that seeding site is 1)
    is_single_source = torch.all(col_sums <= 1).item()

    unique_source_sites = torch.unique(source_sites)

    if len(unique_source_sites) == 0:
        return "no seeding"
    elif is_cyclic(G):
        pattern += "reseeding"
    elif len(unique_source_sites) == 1 or is_single_source:
        pattern += "single-source seeding"
    else:
        pattern += "multi-source seeding"
    return pattern

def get_verbose_seeding_pattern(V, A):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)

    returns: one of: {monoclonal, polyclonal} {primary single-source, single-source, multi-source, reseeding}
    '''
    pattern = get_seeding_pattern(V, A)
    
    G = get_migration_graph(V, A)
    non_zero = torch.where(G > 0)    
    unique_source_sites = torch.unique(non_zero[0])

    if len(unique_source_sites) == 1:
        items = pattern.split(" ")
        items.insert(1, "primary")
        return (" ").join(items)
    return pattern

def get_migration_edges(V, A):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)

    Returns:
        Adjacency matrix where Aij = 1 if there is a migration edge between nodes i and j
    '''
    X = V.T @ V 
    Y = torch.mul(A, (1-X))
    return Y

def get_shared_clusters(V, A, ordered_sites, primary_site, full_node_idx_to_label):
    '''
    returns: list of list of lists with dim: (len(ordered_sites), len(ordered_sites), len(shared_clusters))
    where the innermost list contains the clusters that are shared between site i and site j
    '''
    Y = get_migration_edges(V,A)
    
    shared_clusters = [[[] for x in range(len(ordered_sites))] for y in range(len(ordered_sites))]
    for i,j in tree_iterator(Y):
        site_i = (V[:,i] == 1).nonzero()[0][0].item()
        site_j = (V[:,j] == 1).nonzero()[0][0].item()
        assert(site_i != site_j)
        # if j is a subclonal presence leaf node, add i as the shared cluster 
        # (b/c i is the mutation cluster that j represents)
        if full_node_idx_to_label[j][1] == True:
            shared_clusters[site_i][site_j].append(i)
        else:
            shared_clusters[site_i][site_j].append(j)
    return shared_clusters

def find_highest_level_node(adj_matrix, nodes_to_check):
    start_node = get_root_index(adj_matrix)
    num_nodes = len(adj_matrix)
    visited = [False] * num_nodes

    queue = deque()
    queue.append(start_node)
    visited[start_node] = True

    while queue:
        current_node = queue.popleft()
        if current_node in nodes_to_check:
            return current_node

        for neighbor, connected in enumerate(adj_matrix[current_node]):
            if connected and not visited[neighbor]:
                queue.append(neighbor)
                visited[neighbor] = True
    
def is_monophyletic(adj_matrix, nodes_to_check):
    def dfs(node, target):
        visited[node] = True
        if node == target:
            return True
        for neighbor, connected in enumerate(adj_matrix[node]):
            if connected and not visited[neighbor] and dfs(neighbor, target):
                return True
        return False

    # Initialize variables
    num_nodes = len(adj_matrix)
    visited = [False] * num_nodes
    highest_node = find_highest_level_node(adj_matrix, nodes_to_check)
    if highest_node == get_root_index(adj_matrix):
        return False
    # Check if all nodes can be reached from the top level node in the seeding
    # nodes (seeding node that is closest to the root)
    for node in nodes_to_check:
        visited = [False] * num_nodes
        if not dfs(highest_node, node):
            return False
    return True
    
def get_tracerx_seeding_pattern(V, A, ordered_sites, primary_site, full_node_idx_to_label):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)
    ordered_sites: list of the anatomical site names (e.g. ["breast", "lung_met"]) 
    with length =  num_anatomical_sites) and the order matches the order of cols in V
    primary_site: name of the primary site (must be an element of ordered_sites)

    TRACERx has a different definition of monoclonal vs. polyclonal:
    "If only a single metastatic sample was considered for a case, the case-level 
    dissemination pattern matched the metastasis level dissemination pattern. 
    If multiple metastases were sampled and the dissemination pattern of any 
    individual metastatic sample was defined as polyclonal, the case-level 
    dissemination pattern was also defined as polyclonal. Conversely,if all metastatic 
    samples follow a monoclonal dissemination pattern, all shared clusters between 
    the primary tumour and each metastasis were extracted. If all shared clusters 
    overlapped across all metastatic samples, the case-level dissemination pattern 
    was classified as monoclonal, whereas,  if any metastatic sample shared 
    additional clusters with the primary tumour, the overall dissemination pattern 
    was defined as polyclonal."

    and they define monophyletic vs. polyphyletics as:
    "the origin of the seeding clusters was determined as monophyletic if all 
    clusters appear along a single branch, and polyphyletic if clusters were
    spread across multiple branches of the phylogenetic tree. Thus, if a 
    metastasis was defined as monoclonal, the origin was necessarily monophyletic. 
    For polyclonal metastases, the clusters were mapped to branches of the 
    evolutionary tree. If multiple branches were found, the origin was determined 
    to be polyphyletic, whereas, if only a single branch gave rise to all shared 
    clusters, the origin was defined as monophyletic."
    (from https://www.nature.com/articles/s41586-023-05729-x#Sec7)

    returns: verbal description of the seeding pattern
    '''
    
    Y = get_migration_edges(V,A)
    G = get_migration_graph(V, A)
    non_zero = torch.where(G > 0)
    source_sites = non_zero[0]
    if len(torch.unique(source_sites)) == 0:
        return "no seeding"

    pattern = ""
    # 1) determine if monoclonal (no multi-eges) or polyclonal (multi-edges)
    if len(ordered_sites) == 2:
        pattern = "polyclonal " if ((G > 1).any()) else  "monoclonal "
    elif ((G > 1).any()):
        pattern = "polyclonal "
    else:
        shared_clusters = get_shared_clusters(V, A, ordered_sites, primary_site, full_node_idx_to_label)
        prim_to_met_clusters = shared_clusters[ordered_sites.index(primary_site)]
        all_seeding_clusters = set([cluster for seeding_clusters in prim_to_met_clusters for cluster in seeding_clusters])
        monoclonal = True
        for cluster_set in prim_to_met_clusters:
            # if clusters that seed the primary to each met are not identical,
            # then this is a polyclonal pattern
            if len(cluster_set) != 0 and (set(cluster_set) != all_seeding_clusters):
                monoclonal = False
                break
        pattern = "monoclonal " if monoclonal else "polyclonal "

    # 2) determine if monophyletic or polyphyletic
    if pattern == "monoclonal ":
        pattern += "monophyletic"
        return pattern
    
    seeding_clusters = set()
    for i,j in tree_iterator(Y):
        # if j is a subclonal presence leaf node, add i as the shared cluster 
        # (b/c i is the mutation cluster that j represents)
        if full_node_idx_to_label[j][1] == True:
            seeding_clusters.add(i)
        else:
            seeding_clusters.add(j)
        
    phylo = "monophyletic" if is_monophyletic(A,list(seeding_clusters)) else "polyphyletic"
    
    return pattern + phylo


def write_tree(tree_edge_list, output_filename, add_germline_node=False):
    '''
    Writes the full tree to file like so:
    0 1
    1 2;3
    '''
    if add_germline_node:
        tree_edge_list.append(('GL', tree_edge_list[0][0]))
    with open(output_filename, 'w') as f:
        for edge in tree_edge_list:
            f.write(f"{edge[0]} {edge[1]}")
            f.write("\n")

def write_tree_vertex_labeling(vertex_name_to_site_map, output_filename, add_germline_node=False):
    '''
    Writes the full tree's vertex labeling to file like so:
    1 P
    1_P P
    25;32_M1 M1
    '''
    if add_germline_node:
        vertex_name_to_site_map['GL'] = "P"
    with open(output_filename, 'w') as f:
        for vert_label in vertex_name_to_site_map:
            f.write(f"{vert_label} {vertex_name_to_site_map[vert_label]}")
            f.write("\n")

def write_migration_graph(migration_edge_list, output_filename):
    '''
    Writes the full migration graph to file like so:
    P M1
    P M2
    P M1
    M1 M2
    '''
    with open(output_filename, 'w') as f:
        for edge in migration_edge_list:
            f.write(f"{edge[0]} {edge[1]}")
            f.write("\n")

def plot_losses(losses):
    plt.plot([x for x in range(len(losses))],losses, label="loss")
    plt.xlabel("epoch")
    plt.ylabel("loss")
    plt.show()

def plot_temps(temps):
    plt.plot([x for x in range(len(temps))],temps, label="temp")
    plt.xlabel("epoch")
    plt.ylabel("temp")
    plt.show()

def plot_loss_components(loss_dicts, weights):
    # if else statements to handle lr schedules where we do not calculate
    # all loss components at every epoch

    mig_losses = [e[MIG_KEY] for e in loss_dicts]
    seed_losses = [e[SEEDING_KEY] for e in loss_dicts]
    neg_entropy = [e[ENTROPY_KEY] for e in loss_dicts]

    plt.plot([x for x in range(len(loss_dicts))],mig_losses, label="m")
    plt.plot([x for x in range(len(loss_dicts))],seed_losses, label="s")
    plt.plot([x for x in range(len(loss_dicts))],neg_entropy, label="neg. ent.")

    plt.xlabel("epoch")
    plt.ylabel("loss")
    plt.legend(loc="upper right")
    plt.show()

def view_pydot(pdot):
    plt = Image(pdot.create_png())
    display(plt)

def relabel_cluster(label, shorten, pad):
    if not shorten:
        return label

    out = ""
    # e.g. 1_M2 -> 1_M2
    if len(label) <=4 :
        out = label
    # e.g. 1;3;6;19_M2 -> 1_M2
    elif ";" in label and "_" in label:
        out = label[:label.find(";")] + label[label.find("_"):]
    # e.g. 100_M2 -> 100_M2
    elif "_" in label:
        out = label
    # e.g. 2;14;15 -> 2;14
    else:
        out = ";".join(label.split(";")[:2])
    if pad:
        return out.center(5)
    else:
        return out

def truncated_cluster_name(cluster_name):
    '''
    Displays a max of two mutation names associated with the cluster (e.g. 9;15;19;23;26 -> 9;15)
    Does nothing if the cluster name is not in above format
    '''
    assert(isinstance(cluster_name, str))
    split_name = cluster_name.split(";")
    truncated_name = ";".join(split_name) if len(split_name) <= 2 else ";".join(split_name[:2])
    return truncated_name

def get_full_tree_node_idx_to_label(V, T, custom_node_idx_to_label, ordered_sites, shorten_label=True, pad=False):
    '''
    custom_node_idx_to_label only gives the internal node labels, so build a map of
    node_idx to (label, is_leaf) 
    e.g. ("0;9", False), ("0;9_P", True),  or ("5_liver", True) 
    '''
    full_node_idx_to_label_map = dict()
    for i, j in tree_iterator(T):
        if i in custom_node_idx_to_label:
            full_node_idx_to_label_map[i] = (relabel_cluster(custom_node_idx_to_label[i], shorten_label, pad), False)
        if j in custom_node_idx_to_label:
            full_node_idx_to_label_map[j] = (relabel_cluster(custom_node_idx_to_label[j], shorten_label, pad), False)
        elif j not in custom_node_idx_to_label:
            site_idx = (V[:,j] == 1).nonzero()[0][0].item()
            full_node_idx_to_label_map[j] = (relabel_cluster(f"{custom_node_idx_to_label[i]}_{ordered_sites[site_idx]}", shorten_label, pad), True)
    return full_node_idx_to_label_map

def idx_to_color(custom_colors, idx, alpha=1.0):
    if custom_colors != None:
        rgb = mcolors.to_rgb(custom_colors[idx])
        rgb_alpha = (rgb[0], rgb[1], rgb[2], alpha)
        return mcolors.to_hex(rgb_alpha, keep_alpha=True)

    # TODO repeat colors in this case
    assert(idx < len(COLORS))
    return COLORS[idx]

def get_migration_graph(V, A):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)
    '''
    migration_graph = (V @ A) @ V.T
    migration_graph_no_diag = torch.mul(migration_graph, 1-torch.eye(migration_graph.shape[0], migration_graph.shape[1]))
    
    return migration_graph_no_diag

def plot_migration_graph(V, A, ordered_sites, custom_colors, primary, show=True):
    '''
    Plots migration graph G which represents the migrations/comigrations between
    all anatomical sites.

    Returns a list of edges (e.g. [('P' ,'M1'), ('P', 'M2')])
    '''
    colors = custom_colors
    if colors == None:
        colors = COLORS
    assert(len(ordered_sites) <= len(colors))

    mig_graph_no_diag = get_migration_graph(V, A)

    G = nx.MultiDiGraph()
    for node, color in zip(ordered_sites, colors):
        G.add_node(node, shape="box", color=color, fillcolor='white', fontname="Lato", penwidth=3.0)

    edges = []
    for i, adj_row in enumerate(mig_graph_no_diag):
        for j, num_edges in enumerate(adj_row):
            if num_edges > 0:
                for _ in range(int(num_edges.item())):
                    G.add_edge(ordered_sites[i], ordered_sites[j], color=f'"{colors[i]};0.5:{colors[j]}"', penwidth=3)
                    edges.append((ordered_sites[i], ordered_sites[j]))

    dot = nx.nx_pydot.to_pydot(G)
    if show:
        view_pydot(dot)

    dot_lines = dot.to_string().split("\n")
    dot_lines.insert(1, 'dpi=600;size=3.5;')
    dot_str = ("\n").join(dot_lines)

    return dot_str, edges

def print_averaged_tree(losses_tensor, V, full_trees, node_idx_to_label, custom_colors, ordered_sites, print_config):
    '''
    Returns an averaged tree over all (TODO: all or top k?) converged trees
    by weighing each tree edge or vertex label by the softmax of the 
    likelihood of that tree 
    '''
    _, min_loss_indices = torch.topk(losses_tensor, len(losses_tensor), largest=False, sorted=True)
    # averaged tree edges are weighted by the average of the softmax of the negative log likelihoods
    # TODO*: is this the right way to compute weights?
    def softmaxed_losses(losses_tensor):
        if not torch.is_tensor(losses_tensor):
            losses_tensor = torch.tensor(losses_tensor)
        return torch.softmax(-1.0*(torch.log2(losses_tensor)/ torch.log2(torch.tensor(1.1))), dim=0)

    weights = torch.softmax(-1.0*(torch.log2(losses_tensor)/ torch.log2(torch.tensor(1.1))), dim=0)
    #print("losses tensor\n", losses_tensor, weights)

    weighted_edges = dict() # { edge_0 : [loss_0, loss_1] }
    weighted_node_colors = dict() # { node_0 : { anatomical_site_0 : [loss_0, loss_3]}}
    for sln_idx in min_loss_indices:
        loss = losses_tensor[sln_idx]
        weight = weights[sln_idx]

        full_tree_node_idx_to_label = get_full_tree_node_idx_to_label(V[sln_idx], full_trees[sln_idx], node_idx_to_label, ordered_sites)

        for i, j in tree_iterator(full_trees[sln_idx]):
            edge = full_tree_node_idx_to_label[i][0], full_tree_node_idx_to_label[j][0]
            if edge not in weighted_edges:
                weighted_edges[edge] = []
            weighted_edges[edge].append(weight.item())

        for node_idx in full_tree_node_idx_to_label:
            site_idx = (V[sln_idx][:,node_idx] == 1).nonzero()[0][0].item()
            node_label, _ = full_tree_node_idx_to_label[node_idx]
            if node_label not in weighted_node_colors:
                weighted_node_colors[node_label] = dict()
            if site_idx not in weighted_node_colors[node_label]:
                weighted_node_colors[node_label][site_idx] = []
            weighted_node_colors[node_label][site_idx].append(loss.item())
    
    avg_node_colors = dict()
    for node_label in weighted_node_colors:
        avg_node_colors[node_label] = dict()
        avg_losses = []
        ordered_labels = weighted_node_colors[node_label]
        for site_idx in ordered_labels:
            vals = weighted_node_colors[node_label][site_idx]
            avg_losses.append(sum(vals)/len(vals))

        #softmaxed = np.exp(softmaxed)/sum(np.exp(softmaxed))
        softmaxed = softmaxed_losses(avg_losses)
        for site_idx, soft in zip(ordered_labels, softmaxed):
            avg_node_colors[node_label][site_idx] = soft
    #print("avg_node_colors\n", avg_node_colors)

    avg_edges = dict()
    for edge in weighted_edges:
        avg_edges[edge] = sum(weighted_edges[edge])/len(weighted_edges[edge])

    #print("avg_edges\n", avg_edges)

    plot_averaged_tree(avg_edges, avg_node_colors, ordered_sites, custom_colors, node_idx_to_label, show=print_config.visualize)

# TODO: make custom_node_idx_to_label a required argument

def plot_averaged_tree(avg_edges, avg_node_colors, ordered_sites, custom_colors=None, custom_node_idx_to_label=None, show=True):

    penwidth = 2.0
    alpha = 1.0

    max_edge_weight = max(list(avg_edges.values()))

    def rescaled_edge_weight(edge_weight):
        return (penwidth/max_edge_weight)*edge_weight
    
    
    G = nx.DiGraph()

    for label_i, label_j in avg_edges.keys():
        
        node_i_color = ""
        for site_idx in avg_node_colors[label_i]:
            node_i_color += f'"{idx_to_color(custom_colors, site_idx, alpha=alpha)};{avg_node_colors[label_i][site_idx]}:"'
        node_j_color = ""
        for site_idx in avg_node_colors[label_j]:
            node_j_color += f'"{idx_to_color(custom_colors, site_idx, alpha=alpha)};{avg_node_colors[label_j][site_idx]}:"'
        is_leaf = False

        G.add_node(label_i, xlabel=label_i, label="", shape="circle", fillcolor=node_i_color, 
                    color="none", penwidth=3, style="wedged",
                    fixedsize="true", height=0.35, fontname="Lato", 
                    fontsize="10pt")
        G.add_node(label_j, xlabel="" if is_leaf else label_j, label="", shape="circle", 
                    fillcolor=node_j_color, color="none", 
                    penwidth=3, style="solid" if is_leaf else "wedged",
                    fixedsize="true", height=0.35, fontname="Lato", 
                    fontsize="10pt")

        # G.add_node(label_i, shape="circle", style="wedged", fillcolor=node_i_color, color="none",
        #     alpha=0.5, fontname = "arial", fontsize="10pt", fixedsize="true", width=0.5)
        # G.add_node(label_j, shape="circle", style="wedged", fillcolor=node_j_color, color="none",
        #     alpha=0.5, fontname = "arial", fontsize="10pt", fixedsize="true", width=0.5)
        #print(label_i, label_j, avg_edges[(label_i, label_j)], rescaled_edge_weight(avg_edges[(label_i, label_j)]))
        # G.add_edge(label_i, label_j, color="#black", penwidth=rescaled_edge_weight(avg_edges[(label_i, label_j)]), arrowsize=0.75, spline="ortho")
        style = "dashed" if is_leaf else "solid"
        penwidth = 2 if is_leaf else 2.5
        xlabel = "" if is_leaf else label_j
        G.add_edge(label_i, label_j,
                    color=f'"grey"', 
                    penwidth=rescaled_edge_weight(avg_edges[(label_i, label_j)]), arrowsize=0, fontname="Lato", 
                    fontsize="10pt", style=style)

    #assert(nx.is_tree(G))
    # we have to use graphviz in order to get multi-color edges :/
    dot = to_pydot(G).to_string()
    # hack since there doesn't seem to be API to modify graph attributes...
    dot_lines = dot.split("\n")
    dot_lines.insert(1, 'graph[splines=false]; nodesep=0.7; ranksep=0.6; forcelabels=true;')
    dot = ("\n").join(dot_lines)
    src = Source(dot) # dot is string containing DOT notation of graph
    if show:
        display(src)

def generate_legend_dot(ordered_sites, custom_colors, node_options):
    legend = nx.DiGraph()
    # this whole reversed business is to get the primary at the top of the legend...
    for i, site in enumerate(reversed(ordered_sites)):
        color = idx_to_color(custom_colors, len(ordered_sites)-1-i)
        legend.add_node(i, shape="plaintext", style="solid", label=f"{site}\r", 
                        width=0.3, height=0.2, fixedsize="true",
                        fontname="Lato", fontsize="10pt")
        legend.add_node(f"{i}_circle", fillcolor=color, color=color, 
                        style="filled", height=0.2, **node_options)

    legend_dot = to_pydot(legend).to_string()
    legend_dot = legend_dot.replace("strict digraph", "subgraph cluster_legend")
    legend_dot = legend_dot.split("\n")
    legend_dot.insert(1, 'rankdir="LR";{rank=source;'+" ".join(str(i) for i in range(len(ordered_sites))) +"}")
    legend_dot = ("\n").join(legend_dot)
    return legend_dot


def plot_tree(V, T, gen_dist, ordered_sites, custom_colors=None, custom_node_idx_to_label=None, show=True):

    # (1) Create full directed graph 
    
    # these labels are used for display in plotting
    display_node_idx_to_label_map = get_full_tree_node_idx_to_label(V, T, custom_node_idx_to_label, ordered_sites,
                                                                    shorten_label=True, pad=False)
    # these labels are used for writing out full vertex names to file
    full_node_idx_to_label_map = get_full_tree_node_idx_to_label(V, T, custom_node_idx_to_label, ordered_sites,
                                                                 shorten_label=False, pad=False)

    color_map = { i:idx_to_color(custom_colors, (V[:,i] == 1).nonzero()[0][0].item()) for i in range(V.shape[1])}
    G = nx.DiGraph()
    node_options = {"label":"", "shape": "circle", "penwidth":3, 
                    "fontname":"Lato", "fontsize":"12pt",
                    "fixedsize":"true", "height":0.25}

    # TODO: come up w better scaling mechanism for genetic distance
    if gen_dist != None:
        gen_dist = ((gen_dist / torch.max(gen_dist[gen_dist>0]))*1.5) + 1
        #gen_dist = torch.clamp(gen_dist, max=2) # for visualization purposes

    edges = []
    for i, j in tree_iterator(T):
        label_i, _ = display_node_idx_to_label_map[i]
        label_j, is_leaf = display_node_idx_to_label_map[j]
        
        G.add_node(i, xlabel=label_i, fillcolor=color_map[i], 
                    color=color_map[i], style="filled", **node_options)
        G.add_node(j, xlabel="" if is_leaf else label_j, fillcolor=color_map[j], 
                    color=color_map[j], style="solid" if is_leaf else "filled", **node_options)

        style = "dashed" if is_leaf else "solid"
        penwidth = 5 if is_leaf else 5.5
        xlabel = "" if is_leaf else label_j

        # We're iterating through i,j of the full tree (including leaf nodes),
        # while G only has genetic distances between internal nodes
        minlen = gen_dist[i, j].item() if (gen_dist != None and i < len(gen_dist) and j < len(gen_dist)) else 1.0
        # TODO: make branch lengths a print config option
        #minlen = 1.0
        #print(i, j, minlen)
        G.add_edge(i, j,color=f'"{color_map[i]};0.5:{color_map[j]}"', 
                   penwidth=penwidth, arrowsize=0, style=style, minlen=minlen)

        edges.append((full_node_idx_to_label_map[i][0], full_node_idx_to_label_map[j][0]))

    # Add edge from normal to root 
    root_idx = get_root_index(T)
    root_label = display_node_idx_to_label_map[root_idx][0]
    G.add_node("normal", label="", xlabel=root_label, penwidth=3, style="invis")
    G.add_edge("normal", root_idx, label="", 
                color=f'"{color_map[root_idx]}"', 
                penwidth=4, arrowsize=0, style="solid")

    assert(nx.is_tree(G))

    # (2) Create legend
    #legend_dot = generate_legend_dot(ordered_sites, custom_colors, node_options)

    # we have to use graphviz in order to get multi-color edges :/
    dot = to_pydot(G).to_string().split("\n")
    # hack since there doesn't seem to be API to modify graph attributes...
    dot.insert(1, 'graph[splines=false]; nodesep=0.7; rankdir=TB; ranksep=0.6; forcelabels=true; dpi=600; size=2.5;')
    #dot.insert(2, legend_dot)
    dot_str = ("\n").join(dot)

    if show:
        dot = nx.nx_pydot.to_pydot(dot_str)
        view_pydot(dot)


    vertex_name_to_site_map = { full_node_idx_to_label_map[i][0]:ordered_sites[(V[:,i] == 1).nonzero()[0][0].item()] for i in range(V.shape[1])}
    return dot_str, edges, vertex_name_to_site_map

def get_loss_dict(V, soft_V, T, G, O, p, full_loss):
    V = V.reshape(1, V.shape[0], V.shape[1]) # add batch dimension
    soft_V = soft_V.reshape(1, soft_V.shape[0], soft_V.shape[1]) # add batch dimension
    m, c, s, g, o = vert_label.get_ancestral_labeling_metrics(V, T, G, O, p)
    e = vert_label.get_entropy(V, soft_V)
    loss_dict = {MIG_KEY: m, COMIG_KEY:c, SEEDING_KEY: s, ORGANOTROP_KEY: o, GEN_DIST_KEY: g, ENTROPY_KEY: e}
    loss_dict = {**loss_dict, **{FULL_LOSS_KEY: round(torch.mean(full_loss).item(), 3)}}
    return loss_dict

def convert_lists_to_np_arrays(pickle_outputs, keys):
    '''
    makes unpickling these outputs much faster
    '''
    for key in keys:
        if key in pickle_outputs:
            pickle_outputs[key] = np.array(pickle_outputs[key])

    return pickle_outputs

def print_best_trees(min_loss_solutions, U, ref_matrix, var_matrix, O, weights, ordered_sites,
                     print_config, custom_colors, primary, output_dir, run_name):
    '''
    min_loss_solutions is in order from lowest to highest loss 
    '''
    
    primary_idx = ordered_sites.index(primary)
    p = torch.nn.functional.one_hot(torch.tensor([primary_idx]), num_classes=len(ordered_sites)).T

    ret = None
    figure_outputs = []
    pickle_outputs = {OUT_LABElING_KEY:[], OUT_LOSSES_KEY:[],OUT_IDX_LABEL_KEY:[],
                      OUT_ADJ_KEY:[], OUT_SITES_KEY:ordered_sites, OUT_LOSS_DICT_KEY:[],
                      OUT_PRIMARY_KEY:primary, OUT_SUB_PRES_KEY:U.numpy(), OUT_WEIGHTS_KEY:[],
                      OUT_SOFTV_KEY:[], OUT_GEN_DIST_KEY:[]}

    with torch.no_grad():
        for i, min_loss_solution in enumerate(min_loss_solutions):
            V = min_loss_solution.V
            soft_V = min_loss_solution.soft_V
            T = min_loss_solution.T
            G = min_loss_solution.G
            full_loss = min_loss_solution.loss
            node_idx_to_label = min_loss_solution.node_idx_to_label
            loss_dict = get_loss_dict(V, soft_V, T, G, O, p, full_loss)

            tree_dot, edges, vertices_to_sites_map = plot_tree(V, T, G, ordered_sites, custom_colors, node_idx_to_label, show=False)
            mig_graph_dot, mig_graph_edges = plot_migration_graph(V, T, ordered_sites, custom_colors, primary, show=False)

            seeding_pattern = get_seeding_pattern(V, T)
            figure_outputs.append((tree_dot, mig_graph_dot, loss_dict, seeding_pattern))
            pickle_outputs[OUT_LABElING_KEY].append(V.detach().numpy())
            pickle_outputs[OUT_LOSSES_KEY].append(full_loss.numpy())
            pickle_outputs[OUT_ADJ_KEY].append(T.detach().numpy())
            pickle_outputs[OUT_SOFTV_KEY].append(soft_V.numpy())
            if G != None:
                pickle_outputs[OUT_GEN_DIST_KEY].append(G.numpy())                
            pickle_outputs[OUT_LOSS_DICT_KEY].append(loss_dict)
            pickle_outputs[OUT_WEIGHTS_KEY].append((min_loss_solution.mig_weight, min_loss_solution.comig_weight, min_loss_solution.seed_weight))

            full_tree_idx_to_label = get_full_tree_node_idx_to_label(V, T, node_idx_to_label, ordered_sites,
                                                                     shorten_label=False, pad=False)
            pickle_outputs[OUT_IDX_LABEL_KEY].append(full_tree_idx_to_label)
            if i == 0: # Best tree
                ret = (edges, vertices_to_sites_map, mig_graph_edges, loss_dict)

        #pickle_outputs = convert_lists_to_np_arrays(pickle_outputs, [OUT_LABElING_KEY, OUT_LOSSES_KEY, OUT_ADJ_KEY, OUT_SOFTV_KEY, OUT_GEN_DIST_KEY])

        save_outputs(figure_outputs, print_config, output_dir, run_name, pickle_outputs, weights)

    return ret

def formatted_loss_string(loss_dict, weights):
    s = f"Loss: {loss_dict[FULL_LOSS_KEY]}\n\n"

    s += f"Migration num.: {int(loss_dict[MIG_KEY])}\n"
    s += f"Comigration num.: {int(loss_dict[COMIG_KEY])}\n"
    s += f"Seeding site num.: {int(loss_dict[SEEDING_KEY])}\n"
    s += f"Neg. entropy: {round(float(loss_dict[ENTROPY_KEY]), 3)}\n"

    if weights.gen_dist != 0:
        s += f"Genetic dist. loss: {round(float(loss_dict[GEN_DIST_KEY]), 3)}\n"
    if weights.organotrop != 0:
        s += f"Organotrop. loss: {round(float(loss_dict[ORGANOTROP_KEY]), 3)}\n"
    return s


def save_outputs(figure_outputs, print_config, output_dir, run_name, pickle_outputs, weights):

    if print_config.visualize:
        k = print_config.k_best_trees
        sys_fonts = matplotlib.font_manager.findSystemFonts(fontpaths=None, fontext='ttf')
        for font in sys_fonts:
            if "Lato" in font:
                matplotlib.font_manager.fontManager.addfont(font)
                rcParams['font.family'] = 'Lato'

        n = len(figure_outputs)
        print(run_name)
        if n < k:
            was = "was" if n==1 else "were"
            print(f"{k} unique trees were not found ({n} {was} found). Retry with a higher sample size if you want to get more trees.")
            k = n

        max_trees = 20
        if n > max_trees:
            print("More than 20 solutions detected, only plotting top 20 trees.")
            k = max_trees
        # Create a figure and subplots
        #fig, axs = plt.subplots(3, k*2, figsize=(10, 8))

        plt.suptitle(run_name)

        z = 2 # number of trees displayed per row

        nrows = math.ceil(k/z)
        h = nrows*4
        fig = plt.figure(figsize=(8,h))
        
        vspace = 1/nrows

        for i, (tree_dot, mig_graph_dot, loss_info, seeding_pattern) in enumerate(figure_outputs):
            if i >= max_trees:
                break
            tree = pgv.AGraph(string=tree_dot).draw(format="png", prog="dot", args="-Glabel=\"\"")
            tree = PILImage.open(io.BytesIO(tree))
            mig_graph = pgv.AGraph(string=mig_graph_dot).draw(format="png", prog="dot")
            mig_graph = PILImage.open(io.BytesIO(mig_graph))

            gs1 = gridspec.GridSpec(3, 3)

            row = math.floor(i/2)
            pad = 0.02 if k < 20 else 0.001

            # left = 0.0 if i is odd, 0.55 if even
            # right = 0.45 if i is odd, 1.0 if even
            gs1.update(left=0.0+((i%2)*0.53), right=0.47+0.55*(i%2), top=1-(row*vspace)-pad, bottom=1-((row+1)*vspace)+pad, wspace=0.05)
            ax1 = plt.subplot(gs1[:-1, :])
            ax2 = plt.subplot(gs1[-1, :-1])
            ax3 = plt.subplot(gs1[-1, -1])

            # Render and display each graph
            ax1.imshow(tree)
            ax1.axis('off')
            ax1.set_title(f'Tree {i+1}\n{seeding_pattern}', fontsize=10, loc="left", va="top", x=-0.1, y=1.0)
            #ax1.text(0.1, 0.65, seeding_pattern, fontname=fontname)

            ax2.imshow(mig_graph)
            ax2.axis('off')

            ax3.text(0.5, 0.5, formatted_loss_string(loss_info, weights), ha='center', va='center', fontsize=8)
            ax3.axis('off')

        fig1 = plt.gcf()
        plt.show()
        plt.close()
        if print_config.save_outputs: 
            fig1.savefig(os.path.join(output_dir, f'{run_name}.png'), dpi=600, bbox_inches='tight')

    if print_config.save_outputs:
        if not os.path.isdir(output_dir):
            raise ValueError(f"{output_dir} does not exist.")
        if print_config.verbose: print(f"Saving {run_name} to {output_dir}")
        # Save results to pickle file
        # with open(os.path.join(output_dir, f"{run_name}.pickle"), 'wb') as handle:
        with gzip.open(os.path.join(output_dir,f"{run_name}.pkl.gz"), 'wb') as gzip_file:
            pickle.dump(pickle_outputs, gzip_file, protocol=pickle.HIGHEST_PROTOCOL)
        # Save best dot to file
        tree_dot, mig_graph_dot, _, _ = figure_outputs[0]
        with open(os.path.join(output_dir, f"{run_name}.tree.dot"), 'w') as file:
            file.write(tree_dot)
        with open(os.path.join(output_dir, f"{run_name}.mig_graph.dot"), 'w') as file:
            file.write(mig_graph_dot)



